from __future__ import unicode_literals

from datetime import date
from unittest import skip, skipIf

import django
from django.core.exceptions import (FieldDoesNotExist,
                                    FieldError,
                                    MultipleObjectsReturned,
                                    ObjectDoesNotExist)
from django.db.models import Count
from django.db.models.query import EmptyQuerySet
from django.test import TestCase

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from queryset_sequence import QuerySetSequence

from tests.models import (Article, Author, BlogPost, Book, OnlinePublisher,
                          PeriodicalPublisher, Publisher)


class TestBase(TestCase):
    # The title of each Book, followed by each Article; each ordered by primary
    # key. This matches all without any transforms applied.
    TITLES_BY_PK = [
        'Fiction',
        'Biography',
        'Django Rocks',
        'Alice in Django-land',
        'Some Article',
    ]

    def setUp(self):
        """Set-up some data to be tested against."""
        alice = Author.objects.create(name="Alice")
        bob = Author.objects.create(name="Bob")

        # Purposefully ordered such that the pks will be in the opposite order
        # than the names.
        mad_magazine = PeriodicalPublisher.objects.create(name="Mad Magazine")
        # This publisher is unused, just takes up a PK in the database.
        Publisher.objects.create(name="Unused Publisher")
        big_books = Publisher.objects.create(name="Big Books",
                                             address="123 Street")
        wacky_website = OnlinePublisher.objects.create(name="Wacky Website")

        # Alice wrote some articles.
        Article.objects.create(title="Django Rocks", author=alice,
                               publisher=mad_magazine, release=date(1980, 4, 21))
        Article.objects.create(title="Alice in Django-land", author=alice,
                               publisher=mad_magazine, release=date(1990, 8, 14))

        # Bob wrote a couple of books, an article, and a blog post.
        Book.objects.create(title="Fiction", author=bob, publisher=big_books,
                            pages=10, release=date(2001, 6, 12))
        Book.objects.create(title="Biography", author=bob, publisher=big_books,
                            pages=20, release=date(2002, 12, 24))
        Article.objects.create(title="Some Article", author=bob,
                               publisher=mad_magazine, release=date(1979, 1, 1))
        BlogPost.objects.create(title="Post", author=bob,
                                publisher=wacky_website)

        # Save the authors and publishers for later.
        self.alice = alice
        self.bob = bob
        self.big_books = big_books
        self.mad_magazine = mad_magazine
        self.wacky_website = wacky_website

        # Many tests start with the same QuerySetSequence.
        self.all = QuerySetSequence(Book.objects.all(), Article.objects.all())

        # An empty QuerySetSequence.
        self.empty = QuerySetSequence()


class TestLength(TestBase):
    """
    Ensure that count() and len() are properly summed over the children
    QuerySets.
    """

    def test_count(self):
        # The proper length should be returned via database queries.
        with self.assertNumQueries(2):
            self.assertEqual(self.all.count(), 5)

        # Asking for it again should re-evaluate the query.
        with self.assertNumQueries(2):
            self.assertEqual(self.all.count(), 5)

    def test_len(self):
        # Calling len() evaluates the QuerySet.
        with self.assertNumQueries(2):
            self.assertEqual(len(self.all), 5)

        # Requesting this again should not cause any queries to occur.
        with self.assertNumQueries(0):
            self.assertEqual(len(self.all), 5)

    def test_slice(self):
        """Ensure the proper length is calculated when a slice is taken."""
        with self.assertNumQueries(2):
            self.assertEqual(self.all[1:].count(), 4)

        # This evaluates the QuerySets, which also causes count() on each to be
        # called.
        with self.assertNumQueries(4):
            self.assertEqual(len(self.all[1:]), 4)

    def test_empty_count(self):
        """An empty QuerySetSequence has a count of 0."""
        self.assertEqual(self.empty.count(), 0)

    def test_empty_len(self):
        """An empty QuerySetSequence has a count of 0."""
        self.assertEqual(len(self.empty), 0)


class TestIterator(TestBase):
    """Test the iterator when no ordering is set."""
    def test_iterator(self):
        """Ensure that an iterator queries for the results."""
        # Ensure that calling iterator twice re-evaluates the query.
        with self.assertNumQueries(2):
            data = [it.title for it in self.all.iterator()]
        self.assertEqual(data, TestIterator.TITLES_BY_PK)

        with self.assertNumQueries(2):
            data = [it.title for it in self.all.iterator()]
        self.assertEqual(data, TestIterator.TITLES_BY_PK)

    def test_iter(self):
        """Directly iteratoring the query should return the same results."""
        with self.assertNumQueries(2):
            data = [it.title for it in self.all]
        self.assertEqual(data, TestIterator.TITLES_BY_PK)

    def test_iter_cache(self):
        """Ensure that iterating the QuerySet caches."""
        with self.assertNumQueries(2):
            data = [it.title for it in self.all]
            self.assertEqual(data, TestIterator.TITLES_BY_PK)

        # So the second call does nothing.
        with self.assertNumQueries(0):
            data = [it.title for it in self.all]
            self.assertEqual(data, TestIterator.TITLES_BY_PK)

    def test_empty(self):
        """Test an empty iteration."""
        with self.assertNumQueries(0):
            self.assertEqual(list(self.empty.iterator()), [])

    def test_empty_iter(self):
        with self.assertNumQueries(0):
            self.assertEqual(list(self.empty), [])

    def test_empty_subqueryset(self):
        """Iterating an empty set should work."""
        qss = QuerySetSequence(Book.objects.all(), Article.objects.none()).order_by('title')

        with self.assertNumQueries(1):
            data = [it.title for it in qss]
        self.assertEqual(data, ['Biography', 'Fiction'])


class TestOperators(TestBase):
    def test_and_identity(self):
        """ANDing with an EmptyQuerySet returns an EmptyQuerySet."""
        with self.assertNumQueries(0):
            combined = self.all & BlogPost.objects.none()
        self.assertIsInstance(combined, EmptyQuerySet)

    def test_and(self):
        """ANDing with a QuerySet applies the and to each QuerySet and removes ones of differing types."""
        # ANDing with a different type of QuerySet ends up with an EmptyQuerySet.
        with self.assertNumQueries(0):
            combined = self.all & BlogPost.objects.all()
        self.assertIsInstance(combined, EmptyQuerySet)

        # ANDing with a QuerySet of a type in the QuerySetSequence applies the
        # AND to that QuerySet.
        with self.assertNumQueries(0):
            combined = self.all & Book.objects.filter(pages__lt=15)
        self.assertIsInstance(combined, QuerySetSequence)
        self.assertEqual(len(combined._querysets), 1)
        with self.assertNumQueries(1):
            data = [it.title for it in combined.iterator()]
        self.assertEqual(data, ['Fiction'])

    def test_empty_and(self):
        """An empty QuerySetSequence can be ANDed with a QuerySet, but returns an EmptyQuerySet."""
        combined = self.empty & BlogPost.objects.all()
        self.assertIsInstance(combined, EmptyQuerySet)
        self.assertEqual(list(combined), [])

    def test_or_identity(self):
        """ORing with an empty QuerySet should just return the same QuerySetSequence."""
        with self.assertNumQueries(0):
            combined = self.all | BlogPost.objects.none()
        self.assertIsInstance(combined, QuerySetSequence)
        self.assertEqual(len(combined._querysets), 2)

        with self.assertNumQueries(2):
            data = [it.title for it in combined.iterator()]
        self.assertEqual(data, self.TITLES_BY_PK)

    def test_or(self):
        """ORing with a QuerySet should add it to the list of QuerySets."""
        with self.assertNumQueries(0):
            combined = self.all | BlogPost.objects.all()
        self.assertIsInstance(combined, QuerySetSequence)
        self.assertEqual(len(combined._querysets), 3)

        with self.assertNumQueries(3):
            data = [it.title for it in combined.iterator()]
        self.assertEqual(data, self.TITLES_BY_PK + ['Post'])

    def test_or_querysetsequence(self):
        """ORing with a QuerySetSequence should combine the lists of QuerySets."""
        with self.assertNumQueries(0):
            combined = self.all | QuerySetSequence(BlogPost.objects.all())
        self.assertIsInstance(combined, QuerySetSequence)
        self.assertEqual(len(combined._querysets), 3)

        with self.assertNumQueries(3):
            data = [it.title for it in combined.iterator()]
        self.assertEqual(data, self.TITLES_BY_PK + ['Post'])

    def test_empty_or(self):
        """An empty QuerySetSequence can be ORed with a QuerySet, but returns an EmptyQuerySet."""
        combined = self.empty | BlogPost.objects.all()
        self.assertIsInstance(combined, QuerySetSequence)
        self.assertEqual(len(combined._querysets), 1)

        with self.assertNumQueries(1):
            data = [it.title for it in combined.iterator()]
        self.assertEqual(data, ['Post'])

    def test_empty_or_querysetsequence(self):
        combined = self.empty | QuerySetSequence(BlogPost.objects.all())
        self.assertIsInstance(combined, QuerySetSequence)
        self.assertEqual(len(combined._querysets), 1)

        with self.assertNumQueries(1):
            data = [it.title for it in combined.iterator()]
        self.assertEqual(data, ['Post'])


class TestNone(TestBase):
    def test_none(self):
        """
        Ensure an instance of EmptyQuerySet is returned and has no results (and
        doesn't perform queries).
        """
        with self.assertNumQueries(0):
            qss = self.all.none()
            data = list(qss)

        # This returns an EmptyQuerySet.
        self.assertIsInstance(qss, EmptyQuerySet)

        # Should have no data.
        self.assertEqual(data, [])

    @skip('Currently not working.')
    def test_empty(self):
        qss = self.empty.none()

        # This returns an EmptyQuerySet.
        self.assertIsInstance(qss, EmptyQuerySet)


class TestAll(TestBase):
    def test_all(self):
        """Ensure a copy is made when calling all()."""
        copy = self.all.all()

        # Different QuerySetSequences, but the same content.
        self.assertNotEqual(self.all, copy)

        # Each QuerySet should be evaluated separately.
        with self.assertNumQueries(2):
            data = [it.title for it in self.all]
        self.assertEqual(data, self.TITLES_BY_PK)

        with self.assertNumQueries(2):
            data = [it.title for it in copy]
        self.assertEqual(data, self.TITLES_BY_PK)

    def test_empty(self):
        """Copying an empty QuerySetSequence works fine."""
        copy = self.empty.all()

        # Different QuerySetSequences, but the same content.
        self.assertNotEqual(self.all, copy)
        self.assertEqual(list(copy), [])


class TestRelated(TestBase):
    """Tests for select_related and prefetch_related."""
    # Bob, Bob, Alice, Alice, Bob.
    EXPECTED_ORDER = [2, 2, 1, 1, 2]

    def test_no_related(self):
        """Check the behavior to ensure that iterating causes additional queries."""
        with self.assertNumQueries(2):
            books = list(self.all)
        with self.assertNumQueries(5):
            authors = [b.author.id for b in books]
        self.assertEqual(authors, self.EXPECTED_ORDER)

    def test_select_related(self):
        """Select related removes subsequent queries to get ForeignKeys."""
        with self.assertNumQueries(2):
            books = list(self.all.select_related('author'))
        with self.assertNumQueries(0):
            authors = [b.author.id for b in books]
        self.assertEqual(authors, self.EXPECTED_ORDER)

    # TODO Add a test for select_related that follows multiple ForeignKeys.

    def test_clear_select_related(self):
        """Clearing select related causes all the database queries to occur again."""
        # Ensure there is a database query.
        with self.assertNumQueries(2):
            books = list(self.all.select_related('author').select_related(None))
        with self.assertNumQueries(5):
            authors = [b.author.id for b in books]
        self.assertEqual(authors, self.EXPECTED_ORDER)

    def test_empty_select_related(self):
        """Calling select_related on an empty QuerySetSequence doesn't error."""
        self.empty.select_related('author')

    def test_prefetch_related(self):
        """Now ensure one database query for all authors, per QuerySet."""
        with self.assertNumQueries(4):
            books = list(self.all.prefetch_related('author'))
        with self.assertNumQueries(0):
            authors = [b.author.id for b in books]
        self.assertEqual(authors, self.EXPECTED_ORDER)

    # TODO Add a test for prefetch_related that follows multiple ForeignKeys.

    def test_clear_prefetch_related(self):
        """Ensure the original behavior is restored if prefetch related is cleared."""
        with self.assertNumQueries(2):
            books = list(self.all.prefetch_related('author').prefetch_related(None))
        with self.assertNumQueries(5):
            authors = [b.author.id for b in books]
        self.assertEqual(authors, self.EXPECTED_ORDER)

    def test_empty_prefetch_related(self):
        """Calling prefetch_related on an empty QuerySetSequence doesn't error."""
        self.empty.prefetch_related('author')


class TestDeferOnly(TestBase):
    def test_defer(self):
        """A deferred field will be loaded on access."""
        with self.assertNumQueries(2):
            books = list(self.all.defer('title'))
        with self.assertNumQueries(5):
            titles = [b.title for b in books]
        self.assertEqual(titles, self.TITLES_BY_PK)

    def test_clear_defer(self):
        """Ensure the original behavior is restored if defer is cleared."""
        with self.assertNumQueries(2):
            books = list(self.all.defer('title').defer(None))
        with self.assertNumQueries(0):
            titles = [b.title for b in books]
        self.assertEqual(titles, self.TITLES_BY_PK)

    def test_empty_defer(self):
        """Calling defer on an empty QuerySetSequence doesn't error."""
        self.empty.defer('title')

    def test_only(self):
        """Only causes other fields to load on access (opposite of defer)."""
        with self.assertNumQueries(2):
            books = list(self.all.only('publisher'))
        with self.assertNumQueries(5):
            titles = [b.title for b in books]
        self.assertEqual(titles, self.TITLES_BY_PK)

    # Note that you cannot clear an only call, so None is not a valid value.

    def test_empty_only(self):
        """Calling only on an empty QuerySetSequence doesn't error."""
        self.empty.only('publisher')


class TestUsing(TestBase):
    def test_using(self):
        """Using should be passed through to each QuerySet."""
        with self.assertNumQueries(2):
            titles = [b.title for b in self.all.using('default')]
        self.assertEqual(titles, self.TITLES_BY_PK)


class TestFilter(TestBase):
    def test_filter(self):
        """
        Ensure that filter() properly filters the children QuerySets, note that
        no QuerySets are actually evaluated during this.
        """
        # Filter to just Bob's work.
        with self.assertNumQueries(0):
            bob_qss = self.all.filter(author=self.bob)
        with self.assertNumQueries(2):
            self.assertEqual(bob_qss.count(), 3)

    def test_filter_by_relation(self):
        """
        Ensure that filter() properly filters the children QuerySets when using
        a related model, note that no QuerySets are actually evaluated during
        this.
        """
        # Filter to just Bob's work.
        with self.assertNumQueries(0):
            bob_qss = self.all.filter(author__name=self.bob.name)
        with self.assertNumQueries(2):
            self.assertEqual(bob_qss.count(), 3)

    def test_empty(self):
        """
        Ensure that filter() works when it results in an empty QuerySet.
        """
        # Filter to nothing.
        with self.assertNumQueries(0):
            qss = self.all.filter(title='')
        self.assertIsInstance(qss, QuerySetSequence)
        with self.assertNumQueries(2):
            self.assertEqual(qss.count(), 0)

        # This should not throw an exception.
        data = list(qss)
        self.assertEqual(len(data), 0)

    def _get_qss(self):
        """Returns a QuerySetSequence with 3 QuerySets."""
        return QuerySetSequence(Book.objects.all(),
                                Article.objects.all(),
                                BlogPost.objects.all())

    def test_queryset(self):
        """Test filtering the QuerySets by exact lookups."""
        for key in ['#', '#__exact', '#__iexact']:
            with self.assertNumQueries(0):
                qss = self._get_qss().filter(**{key: 1})

            with self.assertNumQueries(1):
                data = [it.title for it in qss]
            expected = [
                # Just the Articles.
                'Django Rocks',
                'Alice in Django-land',
                'Some Article',
            ]
            self.assertEqual(data, expected)

    def test_queryset_gt(self):
        """Test filtering the QuerySets by > lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().filter(**{'#__gt': 0})

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        expected = [
            # The Articles and BlogPosts.
            'Django Rocks',
            'Alice in Django-land',
            'Some Article',
            'Post',
        ]
        self.assertEqual(data, expected)

        # Additionally this should cast a string input to an int.
        with self.assertNumQueries(0):
            qss = self._get_qss().filter(**{'#__gt': '0'})

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(data, expected)

    def test_queryset_gte(self):
        """Test filtering the QuerySets by >= lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().filter(**{'#__gte': 1})

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        expected = [
            # The Articles and BlogPosts.
            'Django Rocks',
            'Alice in Django-land',
            'Some Article',
            'Post',
        ]
        self.assertEqual(data, expected)

    def test_queryset_lt(self):
        """Test filtering the QuerySets by < lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().filter(**{'#__lt': 2})

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        # The Books and Articles are returned.
        self.assertEqual(data, self.TITLES_BY_PK)

    def test_queryset_lte(self):
        """Test filtering the QuerySets by <= lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().filter(**{'#__lte': 1})

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        # The Books and Articles.
        self.assertEqual(data, self.TITLES_BY_PK)

    def test_queryset_in(self):
        """Filter the QuerySets with the in lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().filter(**{'#__in': [1]})

        with self.assertNumQueries(1):
            data = [it.title for it in qss]
        expected = [
            # Just the Articles.
            'Django Rocks',
            'Alice in Django-land',
            'Some Article',
        ]
        self.assertEqual(data, expected)

    def test_queryset_str(self):
        """Try filtering the QuerySets by various string lookups."""
        for key in ['#__contains', '#__icontains', '#__startswith', '#__istartswith', '#__endswith', '#__iendswith']:
            with self.assertNumQueries(0):
                qss = self._get_qss().filter(**{key: 1})

            with self.assertNumQueries(1):
                data = [it.title for it in qss]
            expected = [
                # Just the Articles.
                'Django Rocks',
                'Alice in Django-land',
                'Some Article',
            ]
            self.assertEqual(data, expected)

    def test_queryset_range(self):
        """Try filtering the QuerySets by the range lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().filter(**{'#__range': [1, 2]})

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        expected = [
            # The Articles and BlogPosts.
            'Django Rocks',
            'Alice in Django-land',
            'Some Article',
            'Post',
        ]
        self.assertEqual(data, expected)

    def test_queryset_unsupported(self):
        """Try filtering the QuerySets by a lookup that doesn't make sense."""
        with self.assertRaises(ValueError):
            self._get_qss().filter(**{'#__year': 1})

    def test_queryset_multiple(self):
        """
        When using multiple paramters to filter they get ANDed together. Ensure
        this works when filtering by QuerySet.
        """
        with self.assertNumQueries(0):
            qss = self._get_qss().filter(**{'#__gt': 0, 'title__gt': 'Django Rocks'})

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        expected = [
            # Some of the Articles and the BlogPosts.
            'Some Article',
            'Post',
        ]
        self.assertEqual(data, expected)

        # This would only look at Articles and BlogPosts, but neither of those
        # have a title > "Some Article."
        qss = self._get_qss().filter(**{'#__gt': 0, 'title__gt': 'Some Article'})

        # Only the articles are here because it's the second queryset.
        data = [it.title for it in qss]
        self.assertEqual(data, [])

    def test_empty(self):
        """Calling filter on an empty QuerySetSequence doesn't error."""
        self.empty.filter(title='')


class TestExclude(TestBase):
    """
    Note that this is the same test as TestFilter, but we exclude the other
    author instead of filtering.
    """

    def test_exclude(self):
        """
        Ensure that exclude() properly filters the children QuerySets, note that
        no QuerySets are actually evaluated during this.
        """
        # Filter to just Bob's work.
        with self.assertNumQueries(0):
            bob_qss = self.all.exclude(author=self.alice)
        with self.assertNumQueries(2):
            self.assertEqual(bob_qss.count(), 3)

    def test_exclude_by_relation(self):
        """
        Ensure that exclude() properly filters the children QuerySets when using
        a related model, note that no QuerySets are actually evaluated during
        this.
        """
        # Filter to just Bob's work.
        with self.assertNumQueries(0):
            bob_qss = self.all.exclude(author__name=self.alice.name)
        with self.assertNumQueries(2):
            self.assertEqual(bob_qss.count(), 3)

    def test_empty(self):
        """
        Ensure that filter() works when it results in an empty QuerySet.
        """
        # Filter to nothing.
        with self.assertNumQueries(0):
            qss = self.all.exclude(author__in=[self.alice, self.bob])
        self.assertIsInstance(qss, QuerySetSequence)
        with self.assertNumQueries(2):
            self.assertEqual(qss.count(), 0)

        # This should not throw an exception.
        data = list(qss)
        self.assertEqual(len(data), 0)

    def _get_qss(self):
        """Returns a QuerySetSequence with 3 QuerySets."""
        return QuerySetSequence(Book.objects.all(),
                                Article.objects.all(),
                                BlogPost.objects.all())

    def test_queryset(self):
        """Test excluding the QuerySets by exact lookups."""
        for key in ['#', '#__exact', '#__iexact']:
            with self.assertNumQueries(0):
                qss = self._get_qss().exclude(**{key: 0})

            # The articles and blogposts are here because the first QuerySet is
            # removed.
            with self.assertNumQueries(2):
                data = [it.title for it in qss]
            expected = [
                # The Articles and BlogPosts.
                'Django Rocks',
                'Alice in Django-land',
                'Some Article',
                'Post',
            ]
            self.assertEqual(data, expected)

    def test_queryset_gt(self):
        """Test excluding the QuerySets by > lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().exclude(**{'#__gt': 1})

        # The books and articles are here.
        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        # The Books and Articles.
        self.assertEqual(data, self.TITLES_BY_PK)

    def test_queryset_gte(self):
        """Test excluding the QuerySets by >= lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().exclude(**{'#__gte': 1})

        # Only the blogposts are here because it's the last queryset.
        with self.assertNumQueries(1):
            data = [it.title for it in qss]
        expected = [
            # Just the Books.
            'Fiction',
            'Biography',
        ]
        self.assertEqual(data, expected)

    def test_queryset_lt(self):
        """Test excluding the QuerySets by < lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().exclude(**{'#__lt': 1})

        # The articles and blogposts are here.
        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        expected = [
            # The Articles and BlogPosts.
            'Django Rocks',
            'Alice in Django-land',
            'Some Article',
            'Post',
        ]
        self.assertEqual(data, expected)

    def test_queryset_lte(self):
        """Test excluding the QuerySets by <= lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().exclude(**{'#__lte': 1})

        # Only the blogposts are here because it's the last queryset.
        with self.assertNumQueries(1):
            data = [it.title for it in qss]
        expected = [
            # The BlogPosts.
            'Post',
        ]
        self.assertEqual(data, expected)

    def test_queryset_in(self):
        """exclude the QuerySets with the in lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().exclude(**{'#__in': [0, 2]})

        # Only the articles are here because it's the second queryset.
        with self.assertNumQueries(1):
            data = [it.title for it in qss]
        expected = [
            # The Articles.
            'Django Rocks',
            'Alice in Django-land',
            'Some Article',
        ]
        self.assertEqual(data, expected)

    def test_queryset_str(self):
        """Try excluding the QuerySets by various string lookups."""
        for key in ['#__contains', '#__icontains', '#__startswith', '#__istartswith', '#__endswith', '#__iendswith']:
            with self.assertNumQueries(0):
                qss = self._get_qss().exclude(**{key: 1})

            # The books and blogposts are here.
            with self.assertNumQueries(2):
                data = [it.title for it in qss]
            expected = [
                # The Books and BlogPosts.
                'Fiction',
                'Biography',
                'Post',
            ]
            self.assertEqual(data, expected)

    def test_queryset_range(self):
        """Try excluding the QuerySets by the range lookup."""
        with self.assertNumQueries(0):
            qss = self._get_qss().exclude(**{'#__range': [1, 2]})

        # Only the books are here as it is the first queryset.
        with self.assertNumQueries(1):
            data = [it.title for it in qss]
        expected = [
            # Just the Articles.
            'Fiction',
            'Biography',
        ]
        self.assertEqual(data, expected)

    def test_queryset_unsupported(self):
        """Try excluding the QuerySets by a lookup that doesn't make sense."""
        with self.assertRaises(ValueError):
            self._get_qss().exclude(**{'#__year': 1})

    def test_empty(self):
        """Calling exclude on an empty QuerySetSequence doesn't error."""
        self.empty.exclude(title='')


class TestExtraAnnotate(TestBase):
    def test_extra(self):
        """Calling extra() gets applied to each QuerySet."""
        # Filter to just Bob's work.
        with self.assertNumQueries(0):
            bob_qss = self.all.extra(where=["author_id = '{}'".format(self.bob.id)])
        with self.assertNumQueries(2):
            self.assertEqual(bob_qss.count(), 3)

    def test_annotate(self):
        """Annotating should get applied to each QuerySet."""
        qss = QuerySetSequence(Publisher.objects.all(), PeriodicalPublisher.objects.all())
        qss = qss.annotate(published_count=Count('published'))

        # Each Published gets the count of things in it added.
        with self.assertNumQueries(2):
            data = [it.published_count for it in qss]
        self.assertEquals(data, [0, 2, 3])


class TestOrderBy(TestBase):
    def test_order_by(self):
        """Ensure that order_by() propagates to QuerySets and iteration."""
        # Order by author and ensure it takes.
        with self.assertNumQueries(0):
            qss = self.all.order_by('title')

        # Check the titles are properly ordered.
        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(data, sorted(self.TITLES_BY_PK))

    def test_order_by_non_existent_field(self):
        """Ordering by a non-existent field raises an exception upon evaluation."""
        with self.assertNumQueries(0):
            qss = self.all.order_by('pages')
        with self.assertRaises(FieldError):
            list(qss)

    def test_order_by_multi(self):
        """Test ordering by multiple fields."""
        # Add another object with the same title, but a later release date.
        Book.objects.create(title="Fiction", author=self.alice,
                            publisher=self.big_books, pages=1,
                            release=date(2018, 10, 3))

        with self.assertNumQueries(0):
            qss = self.all.order_by('title', '-release')

        # Check the titles are properly ordered.
        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        expected = [
            'Alice in Django-land',
            'Biography',
            'Django Rocks',
            'Fiction',
            'Fiction',
            'Some Article',
        ]
        self.assertEqual(data, expected)

    def test_order_by_multi_2(self):
        """Test ordering by multiple fields, where # is not first."""
        # Add another object with the same title, but in a different QuerySet.
        Article.objects.create(title="Fiction", author=self.alice,
                               publisher=self.mad_magazine,
                               release=date(2018, 10, 3))

        with self.assertNumQueries(0):
            qss = self.all.order_by('title', '-#')

        # Check the titles are properly ordered.
        with self.assertNumQueries(2):
            data = [(it.title, it.__class__.__name__) for it in qss]
        expected = [
            ('Alice in Django-land', 'Article'),
            ('Biography', 'Book'),
            ('Django Rocks', 'Article'),
            ('Fiction', 'Article'),
            ('Fiction', 'Book'),
            ('Some Article', 'Article'),
        ]
        self.assertEqual(data, expected)

    def test_order_by_relation(self):
        """
        Apply order_by() with a field that is a relation to another model's id.
        """
        # Order by author and ensure it takes.
        with self.assertNumQueries(0):
            qss = self.all.order_by('author_id')

        # The first two should be Alice, followed by three from Bob.
        with self.assertNumQueries(2):
            for expected, element in zip([self.alice.id] * 2 + [self.bob.id] * 3, qss):
                self.assertEqual(element.author_id, expected)

    def test_order_by_relation_pk(self):
        """
        Apply order_by() with a field that returns a model without a default
        ordering (i.e. using the pk).
        """
        # Order by publisher and ensure it takes.
        with self.assertNumQueries(0):
            qss = self.all.order_by('publisher')

        # Ensure that the test has any hope of passing.
        self.assertLess(self.mad_magazine.pk, self.big_books.pk)

        # The first three should be from Mad Magazine, followed by three from
        # Big Books.
        # Note that the QuerySetSequence itself needs the publisher objects to
        # compare them, so they all get pulled in.
        with self.assertNumQueries(2 + 5):
            for expected, element in zip([self.mad_magazine.id] * 3 + [self.big_books.id] * 2, qss):
                self.assertEqual(element.publisher.id, expected)

    def test_order_by_relation_with_ordering(self):
        """
        Apply order_by() with a field that returns a model with a default
        ordering.
        """
        # Order by author and ensure it takes.
        with self.assertNumQueries(0):
            qss = self.all.order_by('author')

        # The first two should be Alice, followed by three from Bob.
        # Note that the QuerySetSequence itself needs the author objects to
        # compare them, so they all get pulled in.
        with self.assertNumQueries(2 + 5):
            for expected, element in zip([self.alice.id] * 2 + [self.bob.id] * 3, qss):
                self.assertEqual(element.author.id, expected)

    def test_order_by_relation_with_different_ordering(self):
        """
        Apply order_by() with a field that returns a model with different
        ordering on sub-QuerySets.
        """
        # Both of these have publishers with the same fields, but different
        # ordering.
        all = QuerySetSequence(Article.objects.all(), BlogPost.objects.all())

        # Order by publisher and ensure it takes.
        with self.assertNumQueries(0):
            qss = all.order_by('publisher')

        with self.assertRaises(FieldError):
            list(qss)

    def test_order_by_relation_field(self):
        """Apply order_by() with a field through a model relationship."""
        # Order by author name and ensure it takes.
        with self.assertNumQueries(0):
            qss = self.all.order_by('author__name')

        # The first two should be Alice, followed by three from Bob.
        # Note that the QuerySetSequence itself needs the author objects to
        # compare them, so they all get pulled in.
        with self.assertNumQueries(2 + 5):
            for expected, element in zip([self.alice.id] * 2 + [self.bob.id] * 3, qss):
                self.assertEqual(element.author.id, expected)

    def test_order_by_relation_no_existent_field(self):
        """Apply order_by() with a field through a model relationship that doesn't exist."""
        with self.assertNumQueries(0):
            qss = self.all.order_by('publisher__address')

        with self.assertRaises(FieldError):
            list(qss)

    def test_order_by_queryset(self):
        """Ensure we can order by QuerySet and then other fields."""
        # Order by title, but don't interleave each QuerySet.
        with self.assertNumQueries(0):
            qss = self.all.order_by('#', 'title')

        # Ensure that _ordered_iterator isn't called.
        with patch('queryset_sequence.QuerySequenceIterable._ordered_iterator',
                   side_effect=AssertionError('_ordered_iterator should not be called')):
            # Check the titles are properly ordered.
            data = [it.title for it in qss]
            expected = [
                # First the Books, in order.
                'Biography',
                'Fiction',
                # Then the Articles, in order.
                'Alice in Django-land',
                'Django Rocks',
                'Some Article',
            ]
            self.assertEqual(data, expected)

    def test_order_by_queryset_reverse(self):
        """
        It is possible to reverse the order of the internal QuerySets.

        Note that this is *NOT* the same as calling reverse(), as that reverses
        the contents of each QuerySet as well.
        """
        # Order by title, but don't interleave each QuerySet. And reverse
        # QuerySets.
        with self.assertNumQueries(0):
            qss = self.all.order_by('-#', 'title')

        # Ensure that _ordered_iterator isn't called.
        with patch('queryset_sequence.QuerySequenceIterable._ordered_iterator',
                   side_effect=AssertionError('_ordered_iterator should not be called')):
            # Check the titles are properly ordered.
            data = [it.title for it in qss]
            expected = [
                # The articles, in order.
                'Alice in Django-land',
                'Django Rocks',
                'Some Article',
                # Then the Books, in order.
                'Biography',
                'Fiction',
            ]
            self.assertEqual(data, expected)

    def test_empty(self):
        """Calling order_by on an empty QuerySetSequence doesn't error."""
        self.empty.order_by('author')


class TestReverse(TestBase):
    def test_reverse(self):
        """Ensure calling reverse() returns elements in a reverse order."""
        # This really only makes sense if there's an order set.
        with self.assertNumQueries(0):
            qss = self.all.reverse()

        # Note that this didn't really reverse everything because no ordering
        # was set.
        expected = [
            "Django Rocks",
            "Alice in Django-land",
            "Some Article",

            "Fiction",
            "Biography",
        ]

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(data, expected)

    def test_reverse_twice(self):
        """Ensure calling reverse() twice returns elements in a normal order."""
        with self.assertNumQueries(0):
            qss = self.all.reverse().reverse()

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(data, self.TITLES_BY_PK)

    def test_reverse_ordered(self):
        """Reversing an ordered QuerySet should reverse the ordering too."""
        with self.assertNumQueries(0):
            qss = self.all.order_by('title').reverse()

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(data, sorted(self.TITLES_BY_PK, reverse=True))

    def test_reverse_twice_ordered(self):
        """Calling reverse() twice is negated, but then order by title."""
        with self.assertNumQueries(0):
            qss = self.all.reverse().order_by('title').reverse()

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(data, sorted(self.TITLES_BY_PK))

    def test_empty(self):
        """Calling reverse on an empty QuerySetSequence doesn't error."""
        self.empty.reverse()


class TestSlicing(TestBase):
    """Test indexing and slicing."""

    # TODO Caching throughout these test-cases.

    def test_single_element(self):
        """Single element."""
        # 2 counts + evaluating one QuerySet.
        with self.assertNumQueries(3):
            result = self.all[0]
        self.assertEqual(result.title, 'Fiction')
        self.assertIsInstance(result, Book)

    def test_one_QuerySet(self):
        """Test slicing only from one QuerySet."""
        with self.assertNumQueries(0):
            qss = self.all[0:2]

        # 2 counts + evaluating one QuerySet.
        with self.assertNumQueries(3):
            data = [it.title for it in qss]
        self.assertEqual(['Fiction', 'Biography'], data)

    def test_cast_list(self):
        """Test slicing and casting to a list."""
        with self.assertNumQueries(0):
            qss = self.all[0:2]

        # 2 counts + evaluating one QuerySet.
        with self.assertNumQueries(3):
            result = list(qss)
            data = [it.title for it in result]
        self.assertEqual(['Fiction', 'Biography'], data)

    def test_multiple_QuerySets(self):
        """Test slicing across elements from multiple QuerySets."""
        # 2 counts + 2 evaluations.
        with self.assertNumQueries(0):
            qss = self.all[1:3]

        # 2 counts + evaluating two QuerySets.
        with self.assertNumQueries(4):
            data = [it.title for it in qss]
        self.assertEqual(['Biography', 'Django Rocks'], data)

    def test_multiple_slices(self):
        """Test multiple slices taken."""
        with self.assertNumQueries(0):
            result = self.all[1:3]
        self.assertIsInstance(result, QuerySetSequence)
        # Evaluate the QuerySet.
        with self.assertNumQueries(3):
            article = result[1]
        self.assertEqual(article.title, 'Django Rocks')

    def test_multiple_slices_complex(self):
        """Test taking slices multiple times."""
        with self.assertNumQueries(0):
            qss = self.all[1:4]
        self.assertIsInstance(qss, QuerySetSequence)

        with self.assertNumQueries(0):
            qss = qss[1:2]

        # Evaluate the QuerySet.
        with self.assertNumQueries(3):
            data = [it.title for it in qss]
        self.assertEqual(data, ['Django Rocks'])

    def test_step(self):
        """Test behavior when a step is provided to the slice."""
        with self.assertNumQueries(4):
            qss = self.all[0:4:2]
            data = [it.title for it in qss]
        self.assertIsInstance(qss, list)

        self.assertEqual(['Fiction', 'Django Rocks'], data)

    def test_all(self):
        """Test slicing to all elements."""
        with self.assertNumQueries(0):
            qss = self.all[:]
        self.assertIsInstance(qss, QuerySetSequence)

        # Evaluate the QuerySet.
        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(data, self.TITLES_BY_PK)

    def test_slicing_order_by(self):
        """Test slicing when order_by has already been called."""
        # Order by author and take a slice.
        with self.assertNumQueries(0):
            qss = self.all.order_by('title')[1:3]
        self.assertIsInstance(qss, QuerySetSequence)

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(data[0], 'Biography')
        self.assertEqual(data[1], 'Django Rocks')

    def test_open_slice(self):
        """Test slicing without an end."""
        qss = QuerySetSequence(Article.objects.all())[1:]

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(['Alice in Django-land', 'Some Article'], data)

    def test_closed_slice_single_qs(self):
        """Test slicing if the start and end are within the same QuerySet."""
        Article.objects.create(title='Another Article', author=self.bob,
                               publisher=self.mad_magazine,
                               release=date(2018, 10, 3))

        qss = QuerySetSequence(Article.objects.all())[1:3]

        with self.assertNumQueries(2):
            data = [it.title for it in qss]
        self.assertEqual(['Alice in Django-land', 'Some Article'], data)

    def test_empty(self):
        """Slicing an empty QuerySetSequence doesn't error."""
        self.assertEqual(list(self.empty[:]), [])


class TestGet(TestBase):
    def test_get(self):
        """
        Ensure that get() returns the expected element or raises DoesNotExist.
        """
        # Get a particular item.
        with self.assertNumQueries(2):
            book = self.all.get(title='Biography')
        self.assertEqual(book.title, 'Biography')
        self.assertIsInstance(book, Book)

    def test_not_found(self):
        # An exception is rasied if get() is called and nothing is found.
        with self.assertNumQueries(2):
            with self.assertRaises(ObjectDoesNotExist):
                self.all.get(title='')

    def test_multi_found(self):
        """Test multiple found in the same QuerySet."""
        # ...or if get() is called and multiple objects are found.
        with self.assertNumQueries(1):
            with self.assertRaises(MultipleObjectsReturned):
                self.all.get(author=self.bob)

    def test_multi_found_separate_querysets(self):
        """Test one found in each QuerySet."""
        # ...or if get() is called and multiple objects are found.
        with self.assertNumQueries(2):
            with self.assertRaises(MultipleObjectsReturned):
                self.all.get(title__contains='A')

    def test_related_model(self):
        qss = QuerySetSequence(Article.objects.all(), BlogPost.objects.all())
        with self.assertNumQueries(2):
            post = qss.get(publisher__name="Wacky Website")
        self.assertEqual(post.title, 'Post')
        self.assertIsInstance(post, BlogPost)

    def test_queryset_lookup(self):
        """Test using the special QuerySet lookup."""
        with self.assertNumQueries(1):
            article = self.all.get(**{'#': 1, 'author': self.bob})
        self.assertEqual(article.title, 'Some Article')
        self.assertIsInstance(article, Article)

    def test_empty(self):
        """Calling get on an empty QuerySetSequence raises ObjectDoesNotExist."""
        with self.assertRaises(ObjectDoesNotExist):
            self.empty.get(pk=1)


class TestBoolean(TestBase):
    """Tests related to casting the QuerySetSequence to a boolean."""
    def test_exists(self):
        """Ensure that it casts to True if the item is found."""
        with self.assertNumQueries(2):
            self.assertTrue(self.all.filter(title='Biography'))

    def test_exists_second(self):
        """Ensure that it casts to True if the item is found in a subsequent QuerySet."""
        with self.assertNumQueries(2):
            self.assertTrue(self.all.filter(title='Alice in Django-land'))

    def test_not_found(self):
        """Ensure that exists() returns False if the item is not found."""
        with self.assertNumQueries(2):
            self.assertFalse(self.all.filter(title=''))

    def test_multi_found(self):
        """Ensure that it casts to True if multiple items are found."""
        with self.assertNumQueries(2):
            self.assertTrue(self.all.filter(author=self.bob))

    def test_empty(self):
        """An empty QuerySetSequence should resolve to False."""
        self.assertFalse(self.empty)


class TestEarliestLatest(TestBase):
    def test_earliest(self):
        """The earliest release overall should be returned."""
        with self.assertNumQueries(2):
            earliest = self.all.earliest('release')
        self.assertEqual(earliest.title, 'Some Article')

    def test_latest(self):
        """The latest release overall should be returned."""
        with self.assertNumQueries(2):
            latest = self.all.latest('release')
        self.assertEqual(latest.title, 'Biography')

    @skipIf(django.VERSION < (2, 0), 'Support for reversed calls to earliest() and latest() was added in Django 2.0')
    def test_earliest_reverse(self):
        """The earliest release overall reversed is the latest."""
        with self.assertNumQueries(2):
            earliest = self.all.earliest('-release')
        self.assertEqual(earliest.title, 'Biography')

    @skipIf(django.VERSION < (2, 0), 'Support for reversed calls to earliest() and latest() was added in Django 2.0')
    def test_latest_reverse(self):
        """The latest release overall reversed is the earliest."""
        with self.assertNumQueries(2):
            latest = self.all.latest('-release')
        self.assertEqual(latest.title, 'Some Article')

    def test_earliest_get_latest_by(self):
        """Not providing fields causes the get_latest_by field to be used."""
        with self.assertNumQueries(2):
            latest = self.all.earliest()
        self.assertEqual(latest.title, 'Some Article')

    def test_earliest_get_latest_by_error(self):
        """When get_latest_by is used, they must all be the same."""
        with self.assertRaises(ValueError):
            QuerySetSequence(Book.objects.all(), BlogPost.objects.all()).earliest()

    def test_latest_get_latest_by(self):
        """Not providing fields causes the get_latest_by field to be used."""
        with self.assertNumQueries(2):
            latest = self.all.latest()
        self.assertEqual(latest.title, 'Biography')

    def test_latest_get_latest_by_error(self):
        """When get_latest_by is used, they must all be the same."""
        with self.assertRaises(ValueError):
            QuerySetSequence(Book.objects.all(), BlogPost.objects.all()).latest()

    def test_empty(self):
        """An empty QuerySetSequence raises a ValueError."""
        with self.assertRaises(ValueError):
            self.empty.latest()

        with self.assertRaises(ValueError):
            self.empty.earliest()


class TestFirstLast(TestBase):
    def test_first_unordered(self):
        """Should return the first element of the first QuerySet."""
        with self.assertNumQueries(1):
            self.assertEqual(self.all.first().title, 'Fiction')

    def test_last_unordered(self):
        """Should return the last element of the last QuerySet."""
        with self.assertNumQueries(1):
            self.assertEqual(self.all.last().title, 'Some Article')

    def test_empty(self):
        """Empty QuerySetSequence should work."""
        self.assertIsNone(self.empty.first())
        self.assertIsNone(self.empty.last())

    def test_first_ordered(self):
        """When ordering, the items of each QuerySet must be compared."""
        # Order by author and ensure it takes.
        with self.assertNumQueries(2):
            self.assertEqual(self.all.order_by('title').first().title, 'Alice in Django-land')

        with self.assertNumQueries(2):
            self.assertEqual(self.all.order_by('-title').first().title, 'Some Article')

    def test_last_ordered(self):
        """When ordering, the items of each QuerySet must be compared."""
        # Order by author and ensure it takes.
        with self.assertNumQueries(2):
            self.assertEqual(self.all.order_by('-title').last().title, 'Alice in Django-land')

        with self.assertNumQueries(2):
            self.assertEqual(self.all.order_by('title').last().title, 'Some Article')


class TestExists(TestBase):
    def test_exists(self):
        """Ensure that exists() returns True if the item is found in the first QuerySet."""
        with self.assertNumQueries(1):
            self.assertTrue(self.all.filter(title='Biography').exists())

    def test_exists_second(self):
        """Ensure that exists() returns True if the item is found in a subsequent QuerySet."""
        with self.assertNumQueries(2):
            self.assertTrue(self.all.filter(title='Alice in Django-land').exists())

    def test_not_found(self):
        """Ensure that exists() returns False if the item is not found."""
        with self.assertNumQueries(2):
            self.assertFalse(self.all.filter(title='').exists())

    def test_multi_found(self):
        """Ensure that exists() returns True if multiple items are found."""
        with self.assertNumQueries(1):
            self.assertTrue(self.all.filter(author=self.bob).exists())

    def test_empty(self):
        """An empty QuerySetSequence should return False."""
        self.assertFalse(self.empty.exists())


class TestUpdate(TestBase):
    def test_update(self):
        """Update should apply across all QuerySets."""
        # The queries are: creating a save point, the two updates, releasing the
        # save point.
        with self.assertNumQueries(4):
            result = self.all.update(title='A New Title')
        self.assertEqual(result, 5)

        with self.assertNumQueries(2):
            data = [it.title for it in self.all]
        self.assertEqual(data, ['A New Title'] * 5)

    def test_error(self):
        """If an error occurs on one of the QuerySets, no changes should occur."""
        # The queries are: creating a save point, attempting the two updates,
        # releasing the save point.
        with self.assertNumQueries(4):
            with self.assertRaises(FieldDoesNotExist):
                self.all.update(pages=8)

        # The page counts for the Book objects are unmodified.
        with self.assertNumQueries(1):
            data = [it.pages for it in Book.objects.all()]
        self.assertEqual(data, [10, 20])

    def test_empty(self):
        """Calling delete on an empty QuerySetSequence should work."""
        result = self.empty.update()
        self.assertEqual(result, 0)


class TestDelete(TestBase):
    def test_delete_all(self):
        """Ensure that delete() works properly."""
        with self.assertNumQueries(2):
            result = self.all.delete()
        self.assertEqual(result[0], 5)
        self.assertEqual(result[1], {'tests.Article': 3, 'tests.Book': 2})

        with self.assertNumQueries(2):
            self.assertEqual(self.all.count(), 0)

    def test_delete_filter(self):
        """Ensure that delete() works properly when filtering."""
        with self.assertNumQueries(2):
            result = self.all.filter(author=self.alice).delete()
        self.assertEqual(result[0], 2)
        self.assertEqual(result[1], {'tests.Article': 2, 'tests.Book': 0})

        with self.assertNumQueries(2):
            self.assertEqual(self.all.count(), 3)

    def test_empty(self):
        """Calling delete on an empty QuerySetSequence should work."""
        result = self.empty.delete()
        self.assertEqual(result[0], 0)
        self.assertEqual(result[1], {})


class TestExplain(TestBase):
    @skipIf(django.VERSION >= (2, 1), 'explain() added in Django 2.1')
    def test_not_supported(self):
        """Older versions of Django raise an attribute error."""
        with self.assertRaises(AttributeError):
            self.all.explain()

    @skipIf(django.VERSION < (2, 1), 'explain() added in Django 2.1')
    def test_supported(self):
        """Supported versions of Django support explain() and return a string."""
        with self.assertNumQueries(2):
            explanation = self.all.explain()
        # The output of explain is not guaranteed, so do some rough checks.
        self.assertEqual(len(explanation.split('\n')), 2)


class TestGetQueryset(TestBase):
    """Tests related to retrieving QuerySets from the sequence."""
    def test_get_querysets(self):
        """Ensure the correct QuerySet objects are returned."""
        querysets = [Book.objects.all(), Article.objects.all()]
        matched_qss = QuerySetSequence(*querysets)

        self.assertEqual(querysets, matched_qss.get_querysets())


class TestCannotImplement(TestCase):
    """The following methods cannot be implemented in QuerySetSequence."""
    def setUp(self):
        self.all = QuerySetSequence()

    def test_create(self):
        with self.assertRaises(NotImplementedError):
            self.all.create()

    def test_get_or_create(self):
        with self.assertRaises(NotImplementedError):
            self.all.get_or_create()

    def test_update_or_create(self):
        with self.assertRaises(NotImplementedError):
            self.all.update_or_create()

    def test_bulk_create(self):
        with self.assertRaises(NotImplementedError):
            self.all.bulk_create([])

    def test_in_bulk(self):
        with self.assertRaises(NotImplementedError):
            self.all.in_bulk()


class TestNotImplemented(TestCase):
    """The following methods have not been implemented in QuerySetSequence."""
    def setUp(self):
        self.all = QuerySetSequence()

    def test_distinct(self):
        with self.assertRaises(NotImplementedError):
            self.all.distinct()

    def test_values(self):
        with self.assertRaises(NotImplementedError):
            self.all.values()

    def test_values_list(self):
        with self.assertRaises(NotImplementedError):
            self.all.values_list()

    def test_dates(self):
        with self.assertRaises(NotImplementedError):
            self.all.dates(None, None)

    def test_datetimes(self):
        with self.assertRaises(NotImplementedError):
            self.all.datetimes(None, None)

    def test_union(self):
        with self.assertRaises(NotImplementedError):
            self.all.union()

    def test_intersection(self):
        with self.assertRaises(NotImplementedError):
            self.all.intersection()

    def test_difference(self):
        with self.assertRaises(NotImplementedError):
            self.all.difference()

    def test_select_for_update(self):
        with self.assertRaises(NotImplementedError):
            self.all.select_for_update()

    def test_raw(self):
        with self.assertRaises(NotImplementedError):
            self.all.raw()

    def test_aggregate(self):
        with self.assertRaises(NotImplementedError):
            self.all.aggregate()



import codecs

from setuptools import setup, find_packages


def long_description():
    with codecs.open('README.rst', encoding='utf8') as f:
        return f.read()

setup(
    name='django-querysetsequence',
    packages=find_packages(),
    version='0.11dev',
    description='Chain together multiple (disparate) QuerySets to treat them as a single QuerySet.',
    long_description=long_description(),
    author='Percipient Networks, LLC',
    author_email='support@strongarm.io',
    url='https://github.com/percipient/django-querysetsequence',
    download_url='https://github.com/percipient/django-querysetsequence',
    keywords=['django', 'queryset', 'chain', 'multi', 'multiple', 'iterable'],
    license='ISC',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Environment :: Web Environment',
        'Topic :: Internet',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Framework :: Django',
        'License :: OSI Approved :: ISC License (ISCL)',
    ],
    install_requires=[
        'django>=1.11',
    ],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
)



# -*- coding: utf-8 -*-
import django
try:
    from django.contrib.contenttypes.fields import GenericForeignKey
except ImportError:
    from django.contrib.contenttypes.generic import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q
from django.db.models.query import QuerySet

from sys import version_info
from genericm2m import PY3, unicode, str

class GFKOptimizedQuerySet(QuerySet):
    def __init__(self, *args, **kwargs):
        # pop the gfk_field from the kwargs if its passed in explicitly
        self._gfk_field = kwargs.pop('gfk_field', None)

        # call the parent class' initializer
        super(GFKOptimizedQuerySet, self).__init__(*args, **kwargs)

    def _clone(self, *args, **kwargs):
        clone = super(GFKOptimizedQuerySet, self)._clone(*args, **kwargs)
        clone._gfk_field = self._gfk_field
        return clone

    def get_gfk(self):
        if not self._gfk_field:
            for field in self.model._meta.virtual_fields:
                if isinstance(field, GenericForeignKey):
                    self._gfk_field = field
                    break

        return self._gfk_field

    def generic_objects(self, model=None):
        clone = self._clone()

        ctypes_and_fks = {}

        gfk_field = self.get_gfk()
        ctype_field = '%s_id' % gfk_field.ct_field
        fk_field = gfk_field.fk_field

        for obj in clone:
            ctype = ContentType.objects.get_for_id(getattr(obj, ctype_field))
            obj_id = getattr(obj, fk_field)

            ctypes_and_fks.setdefault(ctype, [])
            ctypes_and_fks[ctype].append(obj_id)

        gfk_objects = {}
        for ctype, obj_ids in ctypes_and_fks.items():
            gfk_objects[ctype.pk] = ctype.model_class()._default_manager.in_bulk(obj_ids)

        obj_list = []
        for obj in clone:
            obj = gfk_objects[getattr(obj, ctype_field)][getattr(obj, fk_field)]
            if not model or (model and isinstance(obj, model)):
                obj_list.append(obj)

        return obj_list


class RelatedObjectsDescriptor(object):
    def __init__(self, model=None, from_field='parent', to_field='object'):
        self.related_model = model or RelatedObject
        self.from_field = self.get_related_model_field(from_field)
        self.to_field = self.get_related_model_field(to_field)

    def get_related_model_field(self, field_name):
        opts = self.related_model._meta
        for virtual_field in opts.virtual_fields:
            if virtual_field.name == field_name:
                return virtual_field
        return opts.get_field(field_name)

    def is_gfk(self, field):
        return isinstance(field, GenericForeignKey)

    def get_query_for_field(self, instance, field):
        if self.is_gfk(field):
            ctype = ContentType.objects.get_for_model(instance)
            return {
                field.ct_field: ctype,
                field.fk_field: instance.pk
            }
        elif isinstance(instance, field.rel.to):
            return {field.name: instance}

        raise TypeError(u'Unable to query %s with %s' % (field, instance))

    def get_query_from(self, instance):
        return self.get_query_for_field(instance, self.from_field)

    def get_query_to(self, instance):
        return self.get_query_for_field(instance, self.to_field)

    def contribute_to_class(self, cls, name):
        self.name = name
        self.model_class = cls
        setattr(cls, self.name, self)

    def __get__(self, instance, cls=None):
        if instance is None:
            return self

        ManagerClass = type(self.related_model._default_manager)
        return self.create_manager(instance, ManagerClass)

    def __set__(self, instance, value):
        if instance is None:
            raise AttributeError("Manager must be accessed via instance")

        manager = self.__get__(instance)
        manager.add(*value)

    def delete_manager(self, instance):
        return self.create_manager(instance,
                self.related_model._base_manager.__class__)

    def create_manager(self, instance, superclass, cf_from=True):
        rel_obj = self
        if cf_from:
            core_filters = self.get_query_from(instance)
            rel_field = self.to_field
        else:
            core_filters = self.get_query_to(instance)
            rel_field = self.from_field
        uses_gfk = self.is_gfk(rel_field)

        class RelatedManager(superclass):
            def get_queryset(self):
                if uses_gfk:
                    qs = GFKOptimizedQuerySet(self.model, gfk_field=rel_field)
                    return qs.filter(**(core_filters))
                else:
                    if django.VERSION < (1, 6):
                        method = superclass.get_query_set
                    else:
                        method = superclass.get_queryset

                    return method(self).filter(**(core_filters))

            if django.VERSION < (1, 6):
                get_query_set = get_queryset

            def add(self, *objs):
                for obj in objs:
                    if not isinstance(obj, self.model):
                        raise TypeError(u"'%s' instance expected" % self.model._meta.object_name)
                    if not PY3:
                        for (k, v) in core_filters.iteritems():
                            setattr(obj, k, v)
                    else:
                        for (k, v) in core_filters.items():
                            setattr(obj, k, v)
                    obj.save()
            add.alters_data = True

            def create(self, **kwargs):
                kwargs.update(core_filters)
                return super(RelatedManager, self).create(**kwargs)
            create.alters_data = True

            def get_or_create(self, **kwargs):
                kwargs.update(core_filters)
                return super(RelatedManager, self).get_or_create(**kwargs)
            get_or_create.alters_data = True

            def remove(self, *objs):
                for obj in objs:
                    # Is obj actually part of this descriptor set?
                    if obj in self.all():
                        obj.delete()
                    else:
                        raise rel_obj.related_model.DoesNotExist(
                            u"%r is not related to %r." % (obj, instance))
            remove.alters_data = True

            def clear(self):
                self.all().delete()
            clear.alters_data = True

            def connect(self, obj, **kwargs):
                kwargs.update(rel_obj.get_query_to(obj))
                connection, created = self.get_or_create(**kwargs)
                return connection

            def related_to(self):
                mgr = rel_obj.create_manager(instance, superclass, False)
                return mgr.filter(
                    **rel_obj.get_query_to(instance)
                )

            def symmetrical(self):
                if django.VERSION < (1, 6):
                    method = superclass.get_query_set
                else:
                    method = superclass.get_queryset
                return method(self).filter(
                    Q(**rel_obj.get_query_from(instance)) |
                    Q(**rel_obj.get_query_to(instance))
                ).distinct()

        manager = RelatedManager()
        manager.core_filters = core_filters
        manager.model = self.related_model

        return manager

    def all(self):
        if self.is_gfk(self.from_field):
            ctype = ContentType.objects.get_for_model(self.model_class)
            query = {self.from_field.ct_field: ctype}
        else:
            query = {}
        return self.related_model._default_manager.filter(**query)


class BaseGFKRelatedObject(models.Model):
    """
    A generic many-to-many implementation where diverse objects are related
    across a single model to other diverse objects -> using a dual GFK
    """
    # SOURCE OBJECT:
    parent_type = models.ForeignKey(ContentType, related_name="child_%(class)s")
    parent_id = models.IntegerField(db_index=True)
    parent = GenericForeignKey(ct_field="parent_type", fk_field="parent_id")

    # ACTUAL RELATED OBJECT:
    object_type = models.ForeignKey(ContentType, related_name="related_%(class)s")
    object_id = models.IntegerField(db_index=True)
    object = GenericForeignKey(ct_field="object_type", fk_field="object_id")

    class Meta:
        abstract = True


class RelatedObject(BaseGFKRelatedObject):
    """
    A subclass of BaseGFKRelatedObject which adds two fields used for tracking
    some metadata about the relationship, an alias and the date the relationship
    was created
    """
    alias = models.CharField(max_length=255, blank=True)
    creation_date = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ('-creation_date',)

    def __unicode__(self):
        return unicode(u'%s related to %s ("%s")' % (self.parent, self.object, self.alias))



from __future__ import unicode_literals

from django.db import models
from django.db.models import ForeignKey, CASCADE
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class GroupAbstract(models.Model):

    slug = models.CharField(max_length=128)

    class Meta:
        abstract = True

    def __str__(self):
        return self.slug


@python_2_unicode_compatible
class SectionAbstract(models.Model):

    slug = models.CharField(max_length=128)
    position = models.PositiveIntegerField()

    class Meta:
        abstract = True

    def __str__(self):
        parts = ["%s[%d]" % (self.slug, self.position)]
        if self.group:
            parts.insert(0, "%s" % self.group)
        return "/".join(parts)


@python_2_unicode_compatible
class ItemAbstract(models.Model):

    name = models.CharField(max_length=128)
    position = models.PositiveIntegerField()

    class Meta:
        abstract = True

    def __str__(self):
        parts = ["%s[%d]" % (self.name, self.position)]
        if self.section:
            parts.insert(0, "%s" % self.section)
        return "/".join(parts)


class StackedGroup(GroupAbstract):
    pass


class StackedSection(SectionAbstract):
    group = ForeignKey(StackedGroup, related_name='section_set', on_delete=CASCADE)

    class Meta:
        ordering = ('group', 'position')


class StackedItem(ItemAbstract):
    section = ForeignKey(StackedSection, related_name='item_set', on_delete=CASCADE)

    class Meta:
        ordering = ('section', 'position')


class TabularGroup(GroupAbstract):
    pass


class TabularSection(SectionAbstract):
    group = ForeignKey(TabularGroup, related_name='section_set', on_delete=CASCADE)

    class Meta:
        ordering = ('group', 'position')


class TabularItem(ItemAbstract):
    section = ForeignKey(TabularSection, related_name='item_set', on_delete=CASCADE)

    class Meta:
        ordering = ('section', 'position')


class SortableWithExtraRoot(models.Model):
    slug = models.CharField(max_length=128)


class SortableWithExtraChild(models.Model):
    slug = models.CharField(max_length=128)
    root = ForeignKey(SortableWithExtraRoot, on_delete=CASCADE)
    position = models.PositiveIntegerField()
    foo = models.CharField(max_length=128, default='bar')

    class Meta:
        ordering = ('position', )



from completion import site, DjangoModelProvider

from basic.blog.models import Post
from basic.media.models import Photo
from basic.people.models import Person
from basic.places.models import City, Place


class PostProvider(DjangoModelProvider):
    def get_title(self, obj):
        return obj.title
    
    def get_pub_date(self, obj):
        return obj.publish
    
    def get_data(self, obj):
        return {
            'title': obj.title,
            'url': obj.get_absolute_url(),
        }


class PhotoProvider(DjangoModelProvider):
    def get_title(self, obj):
        return obj.title
    
    def get_pub_date(self, obj):
        return obj.uploaded
    
    def get_data(self, obj):
        return {
            'title': obj.title,
            'url': obj.get_absolute_url(),
        }


class PersonProvider(DjangoModelProvider):
    def get_title(self, obj):
        return obj.full_name
    
    def get_pub_date(self, obj):
        return obj.birth_date
    
    def get_data(self, obj):
        return {
            'title': obj.full_name,
            'url': obj.get_absolute_url(),
        }


class PlaceProvider(DjangoModelProvider):
    def get_title(self, obj):
        return obj.title
    
    def get_pub_date(self, obj):
        return obj.modified
    
    def get_data(self, obj):
        return {
            'title': obj.title,
            'url': obj.get_absolute_url(),
        }


class CityProvider(DjangoModelProvider):
    def get_title(self, obj):
        return unicode(obj)
    
    def get_pub_date(self, obj):
        return None
    
    def get_data(self, obj):
        return {
            'title': unicode(obj),
            'url': obj.get_absolute_url(),
        }


site.register(Post, PostProvider)
site.register(Photo, PhotoProvider)
site.register(Person, PersonProvider)
site.register(Place, PlaceProvider)
site.register(City, CityProvider)



from __future__ import unicode_literals

from django.contrib.contenttypes.fields import GenericRelation, GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import ForeignKey, CASCADE
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class GFKB(models.Model):
    name = models.CharField(max_length=255)
    position = models.PositiveIntegerField()
    content_type = ForeignKey(ContentType, on_delete=CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey()

    class Meta:
        ordering = ['object_id', 'position']

    def __str__(self):
        parts = ["%s[%d]" % (self.name, self.position)]
        if self.content_object:
            parts.insert(0, "%s" % self.content_object)
        return "/".join(parts)


@python_2_unicode_compatible
class GFKA(models.Model):
    slug = models.CharField(max_length=255)
    position = models.PositiveIntegerField()
    content_type = ForeignKey(ContentType, on_delete=CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey()
    b_set = GenericRelation(GFKB)

    class Meta:
        ordering = ['object_id', 'position']

    def __str__(self):
        parts = ["%s[%d]" % (self.slug, self.position)]
        if self.content_object:
            parts.insert(0, "%s" % self.content_object)
        return "/".join(parts)


@python_2_unicode_compatible
class GFKRoot(models.Model):
    slug = models.CharField(max_length=255)
    a_set = GenericRelation(GFKA)

    def __str__(self):
        return self.slug



from genericm2m.models import RelatedObjectsDescriptor


def monkey_patch(model_class, name='related', descriptor=None):
    rel_obj = descriptor or RelatedObjectsDescriptor()
    rel_obj.contribute_to_class(model_class, name)
    setattr(model_class, name, rel_obj)
    return True



"""View for a Select2 widget and QuerySetSequence-based business logic."""
from collections import OrderedDict
from functools import reduce

from dal_queryset_sequence.views import BaseQuerySetSequenceView

from dal_select2.views import Select2ViewMixin

from django.db.models import Q
from django.template.defaultfilters import capfirst
from django.utils import six

from queryset_sequence import QuerySetSequence


class Select2QuerySetSequenceView(BaseQuerySetSequenceView, Select2ViewMixin):
    """
    Combines support QuerySetSequence and Select2 in a single view.

    Example usage::

        url(
            '^your-generic-autocomplete/$',
            autocomplete.Select2QuerySetSequenceView.as_view(
                queryset=autocomplete.QuerySetSequence(
                    Group.objects.all(),
                    TestModel.objects.all(),
                )
            ),
            name='your-generic-autocomplete',
        )

    It is compatible with the :py:mod:`~dal_select2_queryset_sequence.widgets`
    and the fields of :py:mod:`dal_contenttypes`, suits generic relation
    autocompletes.
    """

    def get_results(self, context):
        """
        Return a list of results usable by Select2.

        It will render as a list of one <optgroup> per different content type
        containing a list of one <option> per model.
        """
        groups = OrderedDict()

        for result in context['object_list']:
            groups.setdefault(type(result), [])
            groups[type(result)].append(result)

        return [{
            'id': None,
            'text': capfirst(self.get_model_name(model)),
            'children': [{
                'id': self.get_result_value(result),
                'text': six.text_type(result),
            } for result in results]
        } for model, results in groups.items()]


class Select2QuerySetSequenceAutoView(Select2QuerySetSequenceView):
    """
    Select2QuerySetSequenceAutoView class.

    Filter the queryset based on the models and filter attributes of the
    GenericForeignKeyModelField

    self.model_choice is generated from the Select2GenericForeignKeyModelField,
    see it's docstring
    """

    def get_queryset(self):
        """Return queryset."""
        queryset_models = []
        for model_args in self.model_choice:
            model = model_args[0]
            filter_value = model_args[1]

            kwargs_model = {
                '{}__icontains'.format(filter_value): self.q if self.q else ''
            }
            forward_filtered = [Q(**kwargs_model)]

            try:
                forward_fields = model_args[2]
                for forward in forward_fields:
                    field_key = '{}__icontains'.format(forward[1])
                    field_value = self.forwarded[forward[0]]
                    forward_filtered.append(Q(**{field_key: field_value}))
            except IndexError:
                # if no list on the 3rd index of self.model_choice
                # (reserved for forwarding fields)
                pass

            # link the diffrent field by an & query
            and_forward_filtered = reduce(lambda x, y: x & y, forward_filtered)

            queryset_models.append(model.objects.filter(and_forward_filtered))

        # Aggregate querysets
        qs = QuerySetSequence(*queryset_models)

        # This will limit each queryset so that they show an equal number
        # of results.
        qs = self.mixup_querysets(qs)

        return qs



import textwrap
import json

from django.contrib.staticfiles.storage import staticfiles_storage
try:
    # Django 1.10
    from django.urls import reverse, NoReverseMatch
except ImportError:
    # Django <= 1.9
    from django.core.urlresolvers import reverse, NoReverseMatch
from django.http import HttpResponse, HttpResponseForbidden


def server_data_js(request):
    """
    Dynamic javascript serve defining the grappelli lookup urls.
    """
    if not (request.user.is_active and request.user.is_staff):
        return HttpResponseForbidden('"Permission denied"')

    grappelli_lookup_urls = {}
    # Reverse names are "grp_%(key)s_lookup"
    grappelli_lookup_keys = ("related", "m2m", "autocomplete",)

    for k in grappelli_lookup_keys:
        try:
            grappelli_lookup_urls[k] = reverse("grp_%s_lookup" % k)
        except NoReverseMatch:
            pass

    server_data_js = textwrap.dedent(u"""
        var DJNesting = (typeof window.DJNesting != "undefined")
                       ? DJNesting : {};
        DJNesting.adminStaticPrefix = %s;
        DJNesting.LOOKUP_URLS = %s;""" % (
            json.dumps(staticfiles_storage.url("admin/")),
            json.dumps(grappelli_lookup_urls),))

    return HttpResponse(server_data_js.strip(),
        content_type='application/javascript')



from contextlib import contextmanager
import time
from unittest import expectedFailure, SkipTest

import django
from django.conf import settings
from django.utils import six
from django.utils.text import slugify, unescape_entities

from nested_admin.tests.base import (
    expected_failure_if_grappelli, expected_failure_if_suit,
    skip_if_not_grappelli, BaseNestedAdminTestCase)
from .models import (
    TestAdminWidgetsRoot, TestAdminWidgetsA, TestAdminWidgetsB,
    TestAdminWidgetsC0, TestAdminWidgetsC1)
from .admin import (
    TestAdminWidgetsAInline, TestAdminWidgetsBInline,
    TestAdminWidgetsC0Inline, TestAdminWidgetsC1Inline)


admin_classes = [
    TestAdminWidgetsAInline, TestAdminWidgetsBInline,
    TestAdminWidgetsC0Inline, TestAdminWidgetsC1Inline,
]


@contextmanager
def enable_inline_collapsing():
    """A context manager that configures the inline classes to be collapsible."""
    if 'grappelli' in settings.INSTALLED_APPS:
        class_attr = "inline_classes"
        class_val = ("collapse", "closed", "grp-collapse", "grp-closed")
        reset_val = ("collapse", "open", "grp-collapse", "grp-open")
    else:
        class_attr = "classes"
        class_val = ("collapse", )
        reset_val = None

    for admin in admin_classes:
        setattr(admin, class_attr, class_val)
    try:
        yield
    finally:
        for admin in admin_classes:
            setattr(admin, class_attr, reset_val)


class TestAdminWidgets(BaseNestedAdminTestCase):

    fixtures = ['admin-widgets.xml']

    root_model = TestAdminWidgetsRoot
    nested_models = (TestAdminWidgetsA, TestAdminWidgetsB,
        (TestAdminWidgetsC0, TestAdminWidgetsC1))

    @classmethod
    def setUpClass(cls):
        super(TestAdminWidgets, cls).setUpClass()
        cls.a_model, cls.b_model, (cls.c0_model, cls.c1_model) = cls.nested_models

    def get_name_for_indexes(self, indexes):
        name = "Item %s" % (" ABC"[len(indexes)])
        if name == 'Item C':
            name += "%d%d" % (indexes[-1][0], indexes[-1][1])
        else:
            name += "%d" % indexes[-1]

        name += " (%s)" % " > ".join(["%s" % i[1] for i in self._normalize_indexes(indexes)])
        return name

    def check_prepopulated(self, indexes):
        name = self.get_name_for_indexes(indexes)
        expected_slug = slugify(six.text_type(name))

        slug_sel = self.get_form_field_selector('slug', indexes)

        self.set_field('name', name, indexes)
        time.sleep(0.2)
        slug_val = self.selenium.execute_script(
            'return $("%s").val()' % slug_sel)
        self.assertEqual(slug_val, expected_slug, "prepopulated slug field did not sync")

    def check_datetime(self, indexes):
        date_el = self.get_field('date_0', indexes)
        time_el = self.get_field('date_1', indexes)

        if self.has_grappelli:
            now_link_xpath = "following-sibling::*[1]"
        else:
            now_link_xpath = "following-sibling::*[1]/a[1]"
        date_el.clear()
        time_el.clear()
        date_el.find_element_by_xpath(now_link_xpath).click()
        if self.has_grappelli:
            selector = '#ui-datepicker-div .ui-state-highlight'
            with self.clickable_selector(selector, timeout=1) as el:
                el.click()
        time.sleep(0.1)
        time_el.find_element_by_xpath(now_link_xpath).click()
        if self.has_grappelli:
            selector = '#ui-timepicker .ui-state-active'
            with self.clickable_selector(selector, timeout=1) as el:
                el.click()
        time.sleep(0.10)
        self.assertNotEqual(date_el.get_attribute('value'), '', 'Date was not set')
        self.assertNotEqual(time_el.get_attribute('value'), '', 'Time was not set')

    def check_m2m(self, indexes):
        add_all_link = self.get_field('m2m_add_all_link', indexes)
        remove_all_link = self.get_field('m2m_remove_all_link', indexes)
        remove_all_link.click()
        add_all_link.click()
        m2m_to_sel = self.get_form_field_selector('m2m_to', indexes)
        time.sleep(0.2)
        selected = self.selenium.execute_script(
            'return $("%s").find("option").toArray().map(function(el) { return parseInt(el.value, 10); })'
                % m2m_to_sel)
        self.assertEqual(selected, [1, 2, 3])

    def check_fk(self, indexes):
        field = self.get_field('fk1', indexes)
        parent = field.find_element_by_xpath('parent::*')
        add_related = parent.find_element_by_css_selector('.add-related')
        if self.has_grappelli:
            # Grappelli can be very slow to initialize fk bindings, particularly
            # when run on travis-ci
            time.sleep(1)
        add_related.click()
        name = self.get_name_for_indexes(indexes)
        with self.switch_to_popup_window():
            self.set_field('name', name)
            self.save_form()
        time.sleep(0.1)
        field_id = field.get_attribute('id')
        current_val = self.selenium.execute_script(
            'return $("#%s").find("option:selected").html()' % field_id)
        self.assertEqual(unescape_entities(current_val), name)

    def test_collapsible_inlines(self):
        if not self.has_grappelli and django.VERSION < (1, 10):
            raise SkipTest("Collapsible inlines not supported")

        with enable_inline_collapsing():
            self.load_admin()
            name_field = self.get_field('name', [0])

            self.assertFalse(name_field.is_displayed(), "Inline did not load collapsed")

            if self.has_grappelli:
                collapse_handler = self.selenium.execute_script(
                    'return $(arguments[0]).find("> .djn-collapse-handler")[0]',
                    self.get_item([0]))
            else:
                collapse_handler = self.selenium.execute_script(
                    'return $(arguments[0]).find("> fieldset > h2 > .collapse-toggle")[0]',
                    self.get_group())

            collapse_handler.click()
            self.assertTrue(name_field.is_displayed(), "Inline did not expand")
            collapse_handler.click()
            self.assertFalse(name_field.is_displayed(), "Inline did not collapse")

    def test_initial_extra_prepopulated(self):
        self.load_admin()
        self.check_prepopulated([0])
        self.check_prepopulated([0, 0])

    def test_initial_extra_m2m(self):
        self.load_admin()
        self.check_m2m([0])
        self.check_m2m([0, 0])

    def test_initial_extra_fk_one_deep(self):
        self.load_admin()
        self.check_fk([0])

    def test_initial_extra_fk_two_deep(self):
        self.load_admin()
        if self.has_grappelli:
            time.sleep(0.3)
        self.check_fk([0, 0])

    def test_initial_extra_datetime(self):
        self.load_admin()
        self.check_datetime([0])
        self.check_datetime([0, 0])

    @expected_failure_if_suit  # Known bug with prepopulated fields and django-suit
    def test_add_prepopulated(self):
        self.load_admin()
        self.add_inline()
        self.check_prepopulated([1])

    @expected_failure_if_suit  # Known bug with prepopulated fields and django-suit
    def test_add_initial_extra_prepopulated(self):
        self.load_admin()
        self.add_inline()
        self.check_prepopulated([1, 0])

    def test_add_m2m(self):
        self.load_admin()
        self.add_inline()
        self.check_m2m([1])

    @expected_failure_if_suit  # Known bug with this test, django-suit, and phantomjs
    def test_add_fk(self):
        self.load_admin()
        if self.has_grappelli:
            time.sleep(0.3)
        self.add_inline()
        if self.has_grappelli:
            time.sleep(0.3)
        time.sleep(0.1)
        self.check_fk([1])

    @expectedFailure  # Known bug
    def test_add_initial_extra_m2m(self):
        self.load_admin()
        self.add_inline()
        self.check_m2m([1, 0])

    def test_add_initial_extra_fk(self):
        self.load_admin()
        if self.has_grappelli:
            time.sleep(0.3)
        self.add_inline()
        if self.has_grappelli:
            time.sleep(0.3)
        self.check_fk([1, 0])

    def test_add_datetime(self):
        self.load_admin()
        self.add_inline()
        self.check_datetime([1])

    @expected_failure_if_grappelli  # Known bug with datetime fields and grappelli
    def test_add_initial_extra_datetime(self):
        self.load_admin()
        self.add_inline()
        self.check_datetime([1, 0])

    def test_add_two_deep_m2m(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        self.check_m2m([1, 1])

    def test_add_two_deep_fk(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        self.check_fk([1, 1])

    @expected_failure_if_suit  # Known bug with prepopulated fields and django-suit
    def test_add_two_deep_prepopulated(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        self.check_prepopulated([1, 1])

    @expected_failure_if_grappelli  # Known bug with datetime fields and grappelli
    def test_add_two_deep_datetime(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        self.check_datetime([1, 1])

    def test_add_three_deep_m2m(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        self.add_inline([1, 0, [1]])
        self.check_m2m([1, 0, [1, 0]])

    def test_add_three_deep_fk(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        self.add_inline([1, 0, [1]])
        self.check_fk([1, 0, [1, 0]])

    @expected_failure_if_suit  # Known bug with prepopulated fields and django-suit
    def test_add_three_deep_prepopulated(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        self.add_inline([1, 0, [1]])
        self.check_prepopulated([1, 0, [1, 0]])

    @expected_failure_if_grappelli  # Known bug with datetime fields and grappelli
    def test_add_three_deep_datetime(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        self.add_inline([1, 0, [1]])
        self.check_datetime([1, 0, [1, 0]])

    @skip_if_not_grappelli
    def test_autocomplete_single_init(self):
        self.load_admin()
        self.add_inline()
        self.add_inline([1])
        autocomplete_elements = self.selenium.find_elements_by_xpath(
            '//*[@id="id_testadminwidgetsa_set-1-testadminwidgetsb_set-0-fk2-autocomplete"]')
        self.assertNotEqual(len(autocomplete_elements), 0,
            "Zero autocomplete fields initialized")
        self.assertEqual(len(autocomplete_elements), 1,
            "Too many autocomplete fields initialized")



"""Optionnal shortcuts module for DAL.

This module exports all the public classes for the project. It imports dal
module classes and extension module classes by checking the INSTALLED_APPS
setting:

- if dal_select2 is present, import classes from dal_select2_*,
- with dal_queryset_sequence, import views and fields from it,
- with dal_queryset_sequence and dal_select2, import from
  dal_select2_queryset_sequence,
- with django.contrib.contenttypes, import dal_contenttypes,
- with genericm2m, import dal_genericm2m,
- with gm2m, import dal_gm2m,
- with taggit, import dal_taggit,
- with tagulous, import dal_tagulous.

Note that using this module is optional.
"""

from django.conf import settings as django_settings

from .forms import FutureModelForm
from .views import ViewMixin
from .widgets import (
    Select,
    SelectMultiple,
)


def _installed(*apps):
    for app in apps:
        if app not in django_settings.INSTALLED_APPS:
            return False
    return True


if _installed('dal_select2'):
    from dal_select2.widgets import (
        Select2,
        Select2Multiple,
        ModelSelect2,
        ModelSelect2Multiple,
        TagSelect2,
        ListSelect2
    )
    from dal_select2.views import (
        Select2QuerySetView,
        Select2ListView,
        Select2GroupListView
    )
    from dal_select2.fields import (
        Select2ListChoiceField,
        Select2ListCreateChoiceField
    )

if _installed('dal_queryset_sequence'):
    from dal_queryset_sequence.fields import (
        QuerySetSequenceModelField,
        QuerySetSequenceModelMultipleField,
        GenericForeignKeyModelField,
    )
    from dal_queryset_sequence.views import (
        BaseQuerySetSequenceView,
    )
    from queryset_sequence import QuerySetSequence

if _installed('dal_select2', 'dal_queryset_sequence'):
    from dal_select2_queryset_sequence.views import (
        Select2QuerySetSequenceView,
    )
    from dal_select2_queryset_sequence.widgets import (
        QuerySetSequenceSelect2,
        QuerySetSequenceSelect2Multiple,
    )
    from dal_select2_queryset_sequence.fields import (
        Select2GenericForeignKeyModelField,
    )

if _installed('dal_select2') and _installed('taggit'):
    from dal_select2_taggit.widgets import TaggitSelect2

if _installed('dal_select2') and _installed('tagging'):
    from dal_select2_tagging.widgets import TaggingSelect2

if _installed('genericm2m') and _installed('dal_queryset_sequence'):
    from dal_genericm2m_queryset_sequence.fields import (
        GenericM2MQuerySetSequenceField
    )

if _installed('gm2m') and _installed('dal_queryset_sequence'):
    from dal_gm2m_queryset_sequence.fields import GM2MQuerySetSequenceField



try:
    from distutils.spawn import find_executable
except:
    find_executable = lambda f: None
from datetime import datetime
import inspect
import logging
import os
import shutil
import subprocess
import sys
import tempfile
from unittest import SkipTest

import django
from django.conf import settings
from django.contrib.admin.sites import site as admin_site
from django.test import override_settings
from django.utils.six.moves.urllib.parse import urlparse, urlunparse, ParseResult

from storages.backends.s3boto3 import S3Boto3Storage

from nested_admin.tests.base import (
    BaseNestedAdminTestCase, get_model_name, expected_failure_if_suit)
from .models import (
    PlainStackedRoot, PlainTabularRoot, NestedStackedRoot, NestedTabularRoot)


logger = logging.getLogger(__name__)


def strip_query_from_url(url):
    parts = list(urlparse(url))
    parts[ParseResult._fields.index('query')] = ''
    return urlunparse(parts)


class disable_string_if_invalid_for_grappelli(override_settings):

    def __init__(self):
        self.options = {"TEMPLATES": [settings.TEMPLATES[0].copy()]}
        if 'grappelli' in settings.INSTALLED_APPS:
            self.options['TEMPLATES'][0]['OPTIONS'].pop('string_if_invalid')


@disable_string_if_invalid_for_grappelli()
class VisualComparisonTestCase(BaseNestedAdminTestCase):

    root_model = None
    root_models = [PlainStackedRoot, PlainTabularRoot, NestedStackedRoot, NestedTabularRoot]
    storage = None

    @classmethod
    def setUpClass(cls):
        cls.blinkdiff_bin = os.environ.get('BLINKDIFF_BIN')
        if not cls.blinkdiff_bin:
            cls.blinkdiff_bin = find_executable('blink-diff')
        if not cls.blinkdiff_bin or not os.path.exists(cls.blinkdiff_bin):
            raise SkipTest("blink-diff not installed")
        cls.screenshot_output_dir = os.environ.get('SCREENSHOT_OUTPUT_DIR')
        super(BaseNestedAdminTestCase, cls).setUpClass()
        cls.root_temp_dir = tempfile.mkdtemp()

        if os.environ.get('TRAVIS_BUILD_NUMBER'):
            # For some reason these tests fail on travis when Django > 1.11
            if django.VERSION > (1, 11):
                raise SkipTest("Issue with travis and Django >= 1.11")
            cls.path_prefix = "travis_%s" % os.environ['TRAVIS_BUILD_NUMBER']
        else:
            cls.path_prefix = "local_%s" % datetime.now().strftime('%Y%m%dT%H%M%S')

        if all(k in os.environ for k in ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY']):
            try:
                storage = S3Boto3Storage()
                bucket = storage.bucket  # noqa
            except:
                pass
            else:
                cls.storage = storage

        cls.all_models = {}
        cls.all_model_names = {}

        for root_model in cls.root_models:
            root_admin = admin_site._registry[root_model]

            def descend_admin_inlines(admin):
                data = [admin.model, []]
                for inline in (getattr(admin, 'inlines', None) or []):
                    data[1].append(descend_admin_inlines(inline))
                return data

            cls.all_models[root_model] = models = descend_admin_inlines(root_admin)

            def recursive_map_model_names(data):
                if isinstance(data, list):
                    return [m for m in map(recursive_map_model_names, data)]
                else:
                    return get_model_name(data)

            cls.all_model_names[root_model] = recursive_map_model_names(models)

    @classmethod
    def tearDownClass(cls):
        super(VisualComparisonTestCase, cls).tearDownClass()
        shutil.rmtree(cls.root_temp_dir)

    def setUp(self):
        super(VisualComparisonTestCase, self).setUp()
        self.temp_dir = tempfile.mkdtemp(dir=self.root_temp_dir)
        os.makedirs(os.path.join(self.temp_dir, self.path_prefix))
        self.selenium.set_window_size(780, 600)

    @property
    def models(self):
        return self.all_models[self.root_model]

    @property
    def model_names(self):
        return self.all_model_names[self.root_model]

    def assertSameScreenshot(self, a, b, extra_args=None):
        diff_output_path = a.replace('_a.png', '_diff.png')
        args = [
            self.blinkdiff_bin, "--verbose", "--threshold", "1", "--delta", "0",
            "--output", diff_output_path]

        if self.has_suit:
            suit_left = self.selenium.find_element_by_css_selector('#suit-left')
            args += ['--block-out', "%(x)s,%(y)s,%(w)s,%(h)s" % {
                'x': suit_left.location['x'],
                'y': suit_left.location['y'],
                'w': suit_left.size['width'],
                'h': suit_left.size['height'],
            }]
        if extra_args:
            args += extra_args

        args += [a, b]

        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, _ = p.communicate()
        if p.returncode == 0:
            # No differences found
            if self.screenshot_output_dir:
                os.unlink(a)
                os.unlink(b)
                os.unlink(diff_output_path)
            return
        else:
            logger.info(stdout)
            msg = "Screenshots do not match"
            if self.storage:
                s3_name = "%s/%s" % (self.path_prefix, os.path.basename(diff_output_path))
                with open(diff_output_path, 'rb') as f:
                    s3_name = self.storage.save(s3_name, f)
                s3_url = strip_query_from_url(self.storage.url(s3_name))
                msg = "%s (See <%s>)" % (msg, s3_url)
            elif self.screenshot_output_dir:
                msg = "%s (See %s)" % (msg, diff_output_path)
            raise AssertionError(msg)

    def get_admin_screenshot(self):
        name = inspect.stack()[1][3]
        prefix = "%s/py%s%s_dj%s%s" % (
            (self.path_prefix,) + sys.version_info[:2] + django.VERSION[:2])
        if self.has_grappelli:
            prefix += "_grp"
        elif self.has_suit:
            prefix += "_suit"
        output_dir = self.screenshot_output_dir or self.temp_dir
        suffix = ('a' if self.root_model.__name__.startswith('Plain') else 'b')
        image_path = os.path.join(output_dir, "%s_%s_%s.png" % (prefix, name, suffix))
        self.selenium.save_screenshot(image_path)
        return image_path

    def add_inline(self):
        child_model = self.models[1][0][0]
        verbose_name = child_model._meta.verbose_name.title()
        with self.clickable_xpath('//a[contains(string(.), "Add another %s")]' % verbose_name) as el:
            el.click()

    def test_stacked_empty(self):
        screenshots = []
        for model in [PlainStackedRoot, NestedStackedRoot]:
            self.root_model = model
            self.load_admin()
            screenshots.append(self.get_admin_screenshot())
        self.assertSameScreenshot(*screenshots)

    @expected_failure_if_suit
    def test_tabular_empty(self):
        screenshots = []
        for model in [PlainTabularRoot, NestedTabularRoot]:
            self.root_model = model
            self.load_admin()
            screenshots.append(self.get_admin_screenshot())
        self.assertSameScreenshot(*screenshots)

    @expected_failure_if_suit
    def test_tabular_one_item(self):
        screenshots = []
        for model in [PlainTabularRoot, NestedTabularRoot]:
            self.root_model = model
            child_model = self.models[1][0][0]
            root = model.objects.create(slug='a')
            child_model.objects.create(slug='b', root=root, position=0)
            self.load_admin(obj=root)
            screenshots.append(self.get_admin_screenshot())
        self.assertSameScreenshot(*screenshots)

    def test_stacked_one_item(self):
        screenshots = []
        for model in [PlainStackedRoot, NestedStackedRoot]:
            self.root_model = model
            child_model = self.models[1][0][0]
            root = model.objects.create(slug='a')
            child_model.objects.create(slug='b', root=root, position=0)
            self.load_admin(obj=root)
            screenshots.append(self.get_admin_screenshot())
        self.assertSameScreenshot(*screenshots)

    @expected_failure_if_suit
    def test_tabular_added_item(self):
        screenshots = []
        for model in [PlainTabularRoot, NestedTabularRoot]:
            self.root_model = model
            self.load_admin()
            self.add_inline()
            screenshots.append(self.get_admin_screenshot())
        self.assertSameScreenshot(*screenshots)

    def test_stacked_added_item(self):
        screenshots = []
        for model in [PlainStackedRoot, NestedStackedRoot]:
            self.root_model = model
            self.load_admin()
            self.add_inline()
            screenshots.append(self.get_admin_screenshot())
        self.assertSameScreenshot(*screenshots)

    @expected_failure_if_suit
    def test_tabular_validation_error(self):
        screenshots = []
        for model in [PlainTabularRoot, NestedTabularRoot]:
            self.root_model = model
            self.load_admin()
            if self.has_suit:
                self.selenium.set_window_size(1400, 800)
            self.add_inline()
            with self.clickable_selector('#id_slug') as el:
                el.send_keys('a')
            with self.clickable_selector('#id_children-0-slug') as el:
                el.send_keys('b')
            self.save_form()
            screenshots.append(self.get_admin_screenshot())
        extra_args = []
        if not self.has_grappelli:
            # django has a bug where it doesn't show the 'Remove' link
            # if there is a validationerror on a newly added inline
            # see <https://code.djangoproject.com/ticket/15910>
            delete_col = self.selenium.find_element_by_css_selector('#children-0 .delete')
            extra_args += ['--block-out', "%(x)s,%(y)s,%(w)s,%(h)s" % {
                'x': delete_col.location['x'],
                'y': delete_col.location['y'],
                'w': delete_col.size['width'],
                'h': delete_col.size['height'],
            }]
        self.assertSameScreenshot(*screenshots, extra_args=extra_args)

    @expected_failure_if_suit
    def test_stacked_validation_error(self):
        screenshots = []
        for model in [PlainStackedRoot, NestedStackedRoot]:
            self.root_model = model
            self.load_admin()
            self.add_inline()
            with self.clickable_selector('#id_slug') as el:
                el.send_keys('a')
            with self.clickable_selector('#id_children-0-slug') as el:
                el.send_keys('b')
            self.save_form()
            screenshots.append(self.get_admin_screenshot())
        extra_args = []
        if not self.has_grappelli:
            # django has a bug where it doesn't show the 'Remove' link
            # if there is a validationerror on a newly added inline
            # see <https://code.djangoproject.com/ticket/15910>
            delete_col = self.selenium.find_element_by_css_selector('#children-0 .inline-deletelink')
            extra_args += ['--block-out', "%(x)s,%(y)s,%(w)s,%(h)s" % {
                'x': delete_col.location['x'],
                'y': delete_col.location['y'],
                'w': delete_col.size['width'],
                'h': delete_col.size['height'],
            }]
        self.assertSameScreenshot(*screenshots, extra_args=extra_args)



from django import forms
from django.contrib.contenttypes.models import ContentType
from django.template.defaultfilters import slugify

from basic.blog.models import Post
from basic.media.models import Photo


class BaseRelationshipsForm(forms.ModelForm):
    relationships = forms.CharField(required=False)
    hidden_relationships = forms.CharField(required=False, widget=forms.HiddenInput())
    
    def clean_hidden_relationships(self):
        hidden = self.cleaned_data.get('hidden_relationships') or ''
        
        cts_and_ids = [ct_id for ct_id in hidden.split(',') if ct_id.strip()]
        objects = []
        
        for ct_id in cts_and_ids:
            content_type_id, object_id = ct_id.split(':')
            
            ctype = ContentType.objects.get_for_id(int(content_type_id))
            obj = ctype.model_class()._default_manager.get(pk=object_id)
            
            objects.append(obj)
        
        return objects


class SlugifyMixin(object):
    def clean_title(self):
        title = self.cleaned_data.get('title')
        self.instance.slug = slugify(title)
        return title


class PostForm(BaseRelationshipsForm, SlugifyMixin):
    class Meta:
        model = Post
        fields = ('title', 'body',)


class PhotoForm(BaseRelationshipsForm, SlugifyMixin):
    class Meta:
        model = Photo
        fields = ('title', 'photo',)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-01-01 20:16
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='RelatedObject',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('parent_id', models.IntegerField(db_index=True)),
                ('object_id', models.IntegerField(db_index=True)),
                ('alias', models.CharField(blank=True, max_length=255)),
                ('creation_date', models.DateTimeField(auto_now_add=True)),
                ('object_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='related_relatedobject', to='contenttypes.ContentType')),
                ('parent_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='child_relatedobject', to='contenttypes.ContentType')),
            ],
            options={
                'ordering': ('-creation_date',),
            },
        ),
    ]



from django.conf import settings
from django.contrib import admin

import nested_admin

from .models import (
    PlainStackedRoot, PlainStackedChild, PlainTabularRoot, PlainTabularChild,
    NestedStackedRoot, NestedStackedChild, NestedTabularRoot, NestedTabularChild)


class InlineMixin(object):
    extra = 0
    if 'grappelli' in settings.INSTALLED_APPS:
        sortable_field_name = "position"
    else:
        is_sortable = False
    inline_classes = ("collapse", "open", )
    readonly_fields = ("readonly", )
    suit_classes = ''


class PlainTabularChildInline(InlineMixin, admin.TabularInline):
    model = PlainTabularChild

@admin.register(PlainTabularRoot)
class PlainTabularRootAdmin(admin.ModelAdmin):
    inlines = [PlainTabularChildInline]


class PlainStackedChildInline(InlineMixin, admin.StackedInline):
    model = PlainStackedChild

@admin.register(PlainStackedRoot)
class PlainStackedRootAdmin(admin.ModelAdmin):
    inlines = [PlainStackedChildInline]



class NestedStackedChildInline(InlineMixin, nested_admin.NestedStackedInline):
    model = NestedStackedChild

@admin.register(NestedStackedRoot)
class NestedStackedRootAdmin(nested_admin.NestedModelAdmin):
    inlines = [NestedStackedChildInline]


class NestedTabularChildInline(InlineMixin, nested_admin.NestedTabularInline):
    model = NestedTabularChild

@admin.register(NestedTabularRoot)
class NestedTabularRootAdmin(nested_admin.NestedModelAdmin):
    inlines = [NestedTabularChildInline]



"""Classes for class-based forward declaration."""


class Forward(object):
    """Base class for autocomplete forward declaration."""

    @property
    def type(self):
        """Forward type. Should be implemented in subclasses."""
        raise NotImplementedError("Please use one of my subclasses")

    def to_dict(self):
        """Convert to dictionary which will be rendered as JSON."""
        return {
            "type": self.type
        }


class Field(Forward):
    """Forward field value.

    The type of the forwarded value from the field is either string, list of
    strings or boolean.

    The following rules are used to deduce forwarded type.

     - If there is only one field in the form or subform with name ``src``
    and this field is a checkbox without ``value`` HTML-attribute,
    then boolean value indicating if this checkbox is checked is forwarded.

     - If there is only one field in the form or subform with name ``src``
    and it has ``multiple`` HTML-attribute, then this field is forwarded as a
    list of strings, containing values from this field.

    - If there are one or more fields in the form with name ``src`` and all of
    them are checkboxes with HTML-attribute ``value`` set the list of strings
    containing checked checkboxes is forwarded.

    - Otherwise ``src`` field value forwarded as a string.

    .. py:attribute:: src

        The name of the form field whose value will be forwarded to a view.

    .. py:attribute:: dst

        The name of the key of the forwarded value from the src field in the
        forwarded dictionary. If this value is ``None``, then the key is
        ``src``.
    """

    type = "field"

    def __init__(self, src, dst=None):
        """Instantiate a forwarded field value."""
        self.src = src
        self.dst = dst

    def to_dict(self):
        """Convert to dictionary which will be rendered as JSON."""
        d = super(Field, self).to_dict()

        d.update(src=self.src)
        if self.dst is not None:
            d.update(dst=self.dst)

        return d


class Const(Forward):
    """Forward arbitrary constant value.

    .. py:attribute:: val

        The value to forward. Must be JSON-serializable.

    .. py:attribute:: dst

        The name of the key of the forwarded value.
    """

    type = "const"

    def __init__(self, val, dst):
        """Instantiate a forwarded constant value."""
        self.val = val
        self.dst = dst

    def to_dict(self):
        """Convert to dictionary which will be rendered as JSON."""
        d = super(Const, self).to_dict()

        d.update(val=self.val)
        d.update(dst=self.dst)

        return d


class JavaScript(Forward):
    """Run registered javascript handler and forward its returned value.

    You can register custom forward handler in your JS code as follows:

    .. code-block:: javascript

    yl.registerForwardHandler("your_handler", function (autocompleteElement) {
        // your code here
    });

    Then if your add ``JavaScript("your_handler", "some_value")`` to your
    forwards declaration, your function will be called, autocomplete field
    HTML element will be passed as ``autocompleteElement`` and returned value
    will be added to forward dictionary with ``some_value`` key.

    .. py:attribute:: handler

    The name of the registered handler.

    .. py:attribute:: dst

    The name of the key of the forwarded value from the src field in the
    forwarded dictionary. If this value is ``None``, then the key is
    ``handler``
    """

    type = "javascript"

    def __init__(self, handler, dst=None):
        """Initialize Javascript class."""
        self.handler = handler
        self.dst = dst

    def to_dict(self):
        """Convert to dictionary which will be rendered as JSON."""
        d = super(JavaScript, self).to_dict()

        d.update(handler=self.handler)
        d.update(dst=self.dst)

        return d


class Self(Forward):
    """Forward own value.

    The same as :class:`Field`, except that `src` is always this field
    itself.

    .. py:attribute:: dst

    The name of the key of the forwarded value from the src field in the
    forwarded dictionary. If this value is ``None``, then the key is
    ``self``.
    """

    type = "self"

    def __init__(self, dst=None):
        """Instantiate a forwarded field value."""
        self.dst = dst

    def to_dict(self):
        """Convert to dictionary which will be rendered as JSON."""
        d = super(Self, self).to_dict()

        if self.dst is not None:
            d.update(dst=self.dst)

        return d



"""Widgets for Select2 and django-taggit."""

from dal_select2.widgets import TagSelect2

from django import VERSION
from django.utils import six


class TaggitSelect2(TagSelect2):
    """Select2 tag widget for taggit's TagField."""

    def build_attrs(self, *args, **kwargs):
        """Add data-tags=","."""
        attrs = super(TaggitSelect2, self).build_attrs(*args, **kwargs)
        attrs['data-tags'] = ','
        return attrs

    def value_from_datadict(self, data, files, name):
        """Handle multi-word tag.

        Insure there's a comma when there's only a single multi-word tag,
        or tag "Multi word" would end up as "Multi" and "word".
        """
        value = super(TaggitSelect2, self).value_from_datadict(data,
                                                               files, name)
        if value and ',' not in value:
            value = '%s,' % value
        return value

    def option_value(self, value):
        """Return tag.name attribute of value."""
        return value.tag.name if hasattr(value, 'tag') else value

    def render_options(self, *args):
        """
        Render only selected tags.

        Remove when Django < 1.10 support is dropped.
        """
        selected_choices_arg = 1 if VERSION < (1, 10) else 0
        selected_choices = args[selected_choices_arg]

        # When the data hasn't validated, we get the raw input here
        if isinstance(selected_choices, six.text_type):
            choices = [c.strip() for c in selected_choices.split(',')]
        else:
            # Filter out None values, not needed for autocomplete
            choices = [c.tag.name for c in selected_choices if c]

        options = [
            '<option value="%s" selected="selected">%s</option>' % (
                c, c) for c in choices
        ]

        return '\n'.join(options)



import glob
import os
import tempfile

import django

import django_admin_testutils.settings
import dj_database_url


current_dir = os.path.abspath(os.path.dirname(__file__))
temp_dir = tempfile.mkdtemp()


DEBUG = NESTED_ADMIN_DEBUG = True
DATABASES = {
    'default': dj_database_url.parse(os.environ.get('DATABASE_URL', 'sqlite://:memory:')),
}
SECRET_KEY = 'z-i*xqqn)r0i7leak^#clq6y5j8&tfslp^a4duaywj2$**s*0_'

if django.VERSION > (2, 0):
    MIGRATION_MODULES = {
        'auth': None,
        'contenttypes': None,
        'sessions': None,
    }

try:
    import grappelli  # noqa
except ImportError:
    try:
        import suit  # noqa
    except ImportError:
        INSTALLED_APPS = tuple([])
    else:
        INSTALLED_APPS = tuple(['suit'])
        SUIT_CONFIG = {'CONFIRM_UNSAVED_CHANGES': False}
else:
    INSTALLED_APPS = tuple(['grappelli'])


TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'DIRS': django_admin_testutils.settings.TEMPLATES[0]['DIRS'],
    'APP_DIRS': True,
    'OPTIONS': {
        'string_if_invalid': 'INVALID {{ %s }}',
        'context_processors': [
            'django.contrib.auth.context_processors.auth',
            'django.template.context_processors.debug',
            'django.template.context_processors.i18n',
            'django.template.context_processors.media',
            'django.template.context_processors.static',
            'django.template.context_processors.tz',
            'django.template.context_processors.request',
            'django.contrib.messages.context_processors.messages',
        ],
    },
}]

if 'suit' in INSTALLED_APPS:
    # django-suit has issues with string_if_invalid,
    # so don't use this setting if testing suit.
    TEMPLATES[0]['OPTIONS'].pop('string_if_invalid')


INSTALLED_APPS += (
    'django_admin_testutils',
    'nested_admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.staticfiles',
    'django.contrib.admin',
    'nested_admin.tests',
)

# Add apps within the tests folder
for p in glob.glob(os.path.join(current_dir, '*', 'models.py')):
    INSTALLED_APPS += tuple(["nested_admin.tests.%s" %
        os.path.basename(os.path.dirname(p))])


if django.VERSION >= (1, 10):
    MIDDLEWARE = [
        'django.middleware.security.SecurityMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.common.CommonMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    ]
else:
    MIDDLEWARE_CLASSES = (
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    )

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        'nested_admin.tests': {
            'handlers': ['console'],
            'level': 'WARNING',
        },
    },
}

SITE_ID = 1
ROOT_URLCONF = 'nested_admin.tests.urls'
MEDIA_ROOT = os.path.join(temp_dir, 'media')
MEDIA_URL = '/media/'
STATIC_URL = '/static/'
DEBUG_PROPAGATE_EXCEPTIONS = True
TEST_RUNNER = 'django_admin_testutils.DiscoverRunner'

AWS_S3_REGION_NAME = "us-east-1"
AWS_STORAGE_BUCKET_NAME = 'django-nested-admin-artifacts'



"""User stories, functional tests for AutocompleteTestCase."""
from __future__ import unicode_literals

import time

from django.utils import six

from selenium.common.exceptions import (
    StaleElementReferenceException,
)

import tenacity


class BaseStory(object):
    """Base UserStory class."""

    def __init__(self,
                 case,
                 clear_selector=None,
                 dropdown_selector=None,
                 field_name=None,
                 input_selector=None,
                 label_selector=None,
                 labels_selector=None,
                 model=None,
                 option_selector=None,
                 widget_selector=None):
        """If any kwarg is None, get it from case attributes."""
        self.case = case
        self.clear_selector = clear_selector or self.case.clear_selector
        self.dropdown_selector = (dropdown_selector
                                  or self.case.dropdown_selector)
        self.field_name = field_name or self.case.field_name
        self.input_selector = input_selector or self.case.input_selector
        self.label_selector = label_selector or self.case.label_selector
        self.labels_selector = labels_selector or self.case.labels_selector
        self.model = model or self.case.model
        self.option_selector = option_selector or self.case.option_selector
        self.widget_selector = widget_selector or self.case.widget_selector

        self.field_container_selector = ('fieldset.aligned .field-%s' %
                                         self.field_name)
        self.field_selector = '#id_%s' % self.field_name
        self.field_clear_selector = '%s %s' % (
            self.field_container_selector,
            self.clear_selector
        )
        self.in_popup = False

    @tenacity.retry(stop=tenacity.stop_after_delay(3))
    def find_option(self, text):
        """Incremental sleep until option appeared."""
        options = self.case.browser.find_by_css(self.option_selector)

        for option in options:
            if text in option.text:
                return option

        raise Exception('Option %s not found' % text)

    def get_field_label_selector(self):
        """Return CSS selector for field option label."""
        return '%s %s' % (self.field_container_selector, self.label_selector)

    def clean_label_from_remove_buton(self):
        """Clean child nodes before checking (ie. clear option)."""
        self.case.browser.execute_script(
            '$("%s *").remove()' % self.get_field_label_selector()
        )

    def get_label(self):
        """Return autocomplete widget label."""
        label = self.case.browser.find_by_css(
            self.get_field_label_selector()
        )

        self.clean_label_from_remove_buton()
        return self.clean_label(six.text_type(label.text))

    def clean_label(self, label):
        """Given an option text, return the actual label."""
        return label

    @tenacity.retry(stop=tenacity.stop_after_delay(3))
    def assert_label(self, text):
        """Assert that the autocomplete label matches text."""
        assert six.text_type(text) == six.text_type(self.get_label())

    def get_value(self):
        """Return the autocomplete field value."""
        field = self.case.browser.find_by_css(
            self.field_selector)

        return field['value']

    @tenacity.retry(stop=tenacity.stop_after_delay(3))
    def assert_value(self, value):
        """Assart that the actual field value matches value."""
        assert self.get_value() == six.text_type(value)

    def assert_selection(self, value, label):
        """Assert value is selected and has the given label."""
        self.assert_label(label)
        self.assert_value(value)

    def assert_selection_persists(self, value, label):
        """Assert value and lebel, submit the form and assert again."""
        self.assert_selection(value, label)
        self.submit()
        self.assert_label(label)
        self.assert_value(value)

    @tenacity.retry(stop=tenacity.stop_after_delay(3))
    def assert_suggestion_labels_are(self, expected):
        """Retrying assert that suggestions match expected labels."""
        assert sorted(expected) == sorted(self.get_suggestions_labels())

    def switch_to_popup(self):
        """Wait for and switch to popup window."""
        tries = 10
        while tries:
            if len(self.case.browser.windows) == 2:
                break
            time.sleep(0.1)
            tries -= 1

        self.case.browser.windows.current = self.case.browser.windows[1]
        self.in_popup = True

    def switch_to_main(self):
        """Switch back to main window."""
        self.case.browser.windows.current = self.case.browser.windows[0]
        self.in_popup = False

    def submit(self):
        """Submit the form."""
        sel = 'input[type=submit]'

        if not self.in_popup:
            sel += '[name=_continue]'

        el = self.case.browser.find_by_css(sel).first
        el.click()

        # Wait until the form was actually submited
        tries = 100
        while tries:
            # Popup is gone
            if len(self.case.browser.windows) == 1 and self.in_popup:
                break

            # Page changed
            try:
                el.visible
            except:
                break

            tries -= 1
            time.sleep(.05)

        if not self.in_popup:
            self.case.wait_script()

    def toggle_autocomplete(self):
        """Open the autocomplete dropdown."""
        self.case.click('%s %s' % (
            self.field_container_selector,
            self.widget_selector,
        ))

    def refresh_autocomplete(self):
        """Re-open the autocomplete box."""
        self.toggle_autocomplete()
        self.case.browser.is_element_not_present_by_css(self.option_selector)
        self.toggle_autocomplete()

    def get_suggestions(self):
        """Return the list of suggestions in the autocomplete box."""
        def get_options():
            return self.case.browser.find_by_css(self.option_selector)

        def is_searching(options):
            try:
                return 'Searching' in options[0].text
            except StaleElementReferenceException:
                return True
            except IndexError:
                return True

        options = get_options()
        while is_searching(options) is True:
            time.sleep(0.1)
            options = get_options()
        return [(o['value'], o.text) for o in options]

    def get_suggestions_labels(self):
        """Return labels for suggestions in the autocomplete box."""
        return [o[1] for o in self.get_suggestions()]


class SelectOption(BaseStory):
    """User selects an option."""

    def select_option(self, text):
        """Assert that selecting option "text" sets input's value."""
        dropdown = self.case.browser.find_by_css(self.dropdown_selector)
        if not len(dropdown) or not dropdown.visible:
            self.toggle_autocomplete()

        self.case.assert_visible(self.dropdown_selector)
        self.case.enter_text(self.input_selector, text)
        self.find_option(text).click()

    def clear_option(self):
        """Clear current option."""
        self.case.click(self.field_clear_selector)


class InlineSelectOption(SelectOption):
    """Same as UserCanSelectOption, in a given inline."""

    def __init__(self, case, inline_number, inline_related_name=None,
                 **kwargs):
        """
        Do the same as UserCanSelectOption, in inline inline_number.

        Where inline_related_name should be the related_name option for the
        foreign key used for the InlineModelAdmin.
        """
        self.inline_number = inline_number
        self.inline_related_name = (inline_related_name
                                    or case.inline_related_name)

        super(InlineSelectOption, self).__init__(case, **kwargs)

        self.field_container_selector = '#%s-%s .field-%s' % (
            self.inline_related_name, self.inline_number, self.field_name)
        self.field_selector = '#id_%s-%s-%s' % (
            self.inline_related_name,
            self.inline_number,
            self.field_name
        )

        # Ensure the inline is displayed else click to add it
        add = self.case.browser.find_link_by_partial_text('Add another').first

        num = len(
            self.case.browser.find_by_css(
                '.dynamic-%s' % self.inline_related_name
            )
        )
        while num < self.inline_number + 1:
            add.click()

            # Did it work or wasn't the js loaded yet ?
            try:
                self.case.browser.find_by_css(
                    self.field_selector.replace(
                        str(self.inline_number),
                        str(num),
                    )
                ).first  # as usual, rely on implicit wait
            except:
                continue

            num += 1


class RenameOption(SelectOption):
    """User story to rename an option in django admin."""

    def rename_option(self, current_name, add_keys):
        """Click the change button and rename in the popup."""
        self.case.click('#change_id_%s' % self.field_name)

        self.switch_to_popup()
        name_input = self.case.browser.find_by_id('id_name')
        self.case.assertEquals(
            name_input['value'],
            current_name
        )

        self.case.enter_text('#id_name', add_keys)
        self.submit()

        self.switch_to_main()


class AddAnotherOption(BaseStory):
    """Add-another user story."""

    def add_another(self, name):
        """Click the add button and add another option in the popup."""
        self.case.browser.find_by_id('add_id_%s' % self.field_name).click()

        self.switch_to_popup()
        self.case.enter_text('#id_name', name)
        self.submit()
        self.switch_to_main()


class CreateOption(SelectOption):
    """Create an option on the fly."""

    def create_option(self, name):
        """
        Select the only option after typing name and submit.

        name should be unique.
        """
        self.toggle_autocomplete()
        self.case.enter_text(self.input_selector, name)

        self.case.browser.is_element_present_by_text(name)
        self.case.click(self.option_selector)
        self.case.browser.is_element_not_present_by_css(
            '.select2-results__options'
        )


class MultipleMixin(object):
    """Enable multiple choice support with stories."""

    def get_field_labels_selector(self):
        """Return CSS selector for field option label."""
        return '%s %s' % (self.field_container_selector, self.labels_selector)

    def clean_label_from_remove_buton(self):
        """Clean child nodes before checking (ie. clear option)."""
        self.case.browser.execute_script(
            '$("%s span").remove()' % self.get_field_labels_selector()
        )

    def get_labels(self):
        """Return autocomplete widget label."""
        self.clean_label_from_remove_buton()

        labels = self.case.browser.find_by_css(
            self.get_field_labels_selector()
        )

        return [
            self.clean_label(six.text_type(label.text))
            for label in labels
        ]

    def get_values(self):
        """Return the autocomplete field value."""
        script = """
        window.GET_VALUES = [];
        $('%s option:selected').each(function() {
            GET_VALUES.push($(this).attr('value'));
        });
        """ % self.field_selector
        self.case.browser.execute_script(script)
        return self.case.browser.evaluate_script('window.GET_VALUES')

    def assert_labels(self, texts):
        """Assert that all labels match texts."""
        labels = self.get_labels()

        for text in texts:
            self.case.assertIn(text, labels)

        self.case.assertEquals(len(texts), len(labels))

    def assert_values(self, values):
        """Assart that the actual field values matches values."""
        text_values = [six.text_type(v) for v in values]
        actual_values = self.get_values()

        for actual_value in actual_values:
            self.case.assertIn(
                actual_value,
                text_values
            )

        self.case.assertEquals(len(values), len(actual_values))

    def assert_selection(self, values, labels):
        """Assert selections have values and labels."""
        self.assert_labels(labels)
        self.assert_values(values)

    def assert_selection_persists(self, values, labels):
        """Do the same as above, but also submits the form and check again."""
        self.assert_selection(values, labels)
        self.submit()
        self.assert_labels(labels)
        self.assert_values(values)


class CreateOptionMultiple(MultipleMixin, CreateOption):
    """Multiple version of CreateOptions."""


class SelectOptionMultiple(MultipleMixin, SelectOption):
    """Multiple version of CreateOptions."""


class InlineSelectOptionMultiple(MultipleMixin, InlineSelectOption):
    """Multiple options for InlineSelectOption."""

    def __init__(self, case, inline_number, inline_related_name=None,
                 **kwargs):
        """Set input_selector with field_container_selector."""
        super(InlineSelectOptionMultiple, self).__init__(
            case,
            inline_number,
            inline_related_name=inline_related_name,
            **kwargs
        )

        self.input_selector = '%s %s' % (
            self.field_container_selector,
            self.input_selector,
        )



#!/usr/bin/env python
import os
import re
import warnings

import django_admin_testutils


# A dict for mapping test case classes to their import paths, to allow passing
# TestCaseClass.test_function as shorthand to runtests.py
TEST_CASE_MODULE_PATHS = {
    'TestAdminWidgets': 'nested_admin.tests.admin_widgets.tests',
    'TestGenericInlineAdmin': 'nested_admin.tests.gfk.tests',
    'VisualComparisonTestCase': 'nested_admin.tests.one_deep.tests',
    'TestDeepNesting': 'nested_admin.tests.three_deep.tests',
    'TestStackedInlineAdmin': 'nested_admin.tests.two_deep.tests',
    'TestTabularInlineAdmin': 'nested_admin.tests.two_deep.tests',
    'TestSortablesWithExtra': 'nested_admin.tests.two_deep.tests',
    'TestIdenticalPrefixes': 'nested_admin.tests.identical_prefixes.tests',
}


def expand_test_module(module):
    module = os.path.normpath(module)
    matches = re.search(r'^([^/.]+)(\.[^./]+)?$', module)
    if not matches:
        return module
    cls, test_fn = matches.groups()
    if not test_fn:
        test_fn = ''
    if cls not in TEST_CASE_MODULE_PATHS:
        return module
    return "%s.%s%s" % (TEST_CASE_MODULE_PATHS[cls], cls, test_fn)


class RunTests(django_admin_testutils.RunTests):

    def execute(self, flags, test_labels):
        test_labels = [expand_test_module(m) for m in test_labels]
        super(RunTests, self).execute(flags, test_labels)


def main():
    warnings.simplefilter("error", Warning)
    warnings.filterwarnings(
        "ignore",
        "name used for saved screenshot does not match file type",
        UserWarning)
    runtests = RunTests("nested_admin.tests.settings", "nested_admin")
    runtests()


if __name__ == '__main__':
    main()



import django
from django.conf import settings
from django.conf.urls import include, url
from django.contrib import admin

# Explicitly import to register the admins for the test models
for app in settings.INSTALLED_APPS:
    if app.startswith('nested_admin.tests.'):
        __import__("%s.admin" % app)


urlpatterns = [
    url(r'^_nesting/', include('nested_admin.urls')),
]

if django.VERSION < (1, 9):
    urlpatterns += [url(r'^admin/', include(admin.site.urls))]
else:
    urlpatterns += [url(r'^admin/', admin.site.urls)]

try:
    import grappelli
except ImportError:
    pass
else:
    urlpatterns += [url(r"^grappelli/", include("grappelli.urls"))]



from providers import *

from basic.blog.models import Post
from basic.media.models import Photo
from completion.listeners import start_listening
from genericm2m.utils import monkey_patch


# monkey patch the Post model with a related objects descriptor
monkey_patch(Post)
monkey_patch(Photo)

# configure our signal handlers so we can update the autocomplete index on
# model save & delete
start_listening()



"""
More efficient pagination for django-querysetsequence when using Django REST
Framework.

The standard Django REST Framework pagination classes will work fine. The
classes provided here are useful when providing large datasets that use
QuerySetSequence over an API as they first order by the QuerySet number and then
particular fields.

"""
from base64 import b64decode

from django.utils.six.moves.urllib import parse as urlparse

from queryset_sequence import QuerySetSequence

try:
    from rest_framework.exceptions import NotFound
    from rest_framework.pagination import (_positive_int,
                                           _reverse_ordering,
                                           Cursor,
                                           CursorPagination)
except ImportError:
    # This requires Django REST Framework to be installed.
    raise ImportError(
        "queryset_sequence.pagination is for use with Django REST Framework, "
        "which was not found.")


class SequenceCursorPagination(CursorPagination):
    """
    Heavily customized CursorPagination to first sort by QuerySet, then by a
    different ordering field.

    Changes include:
    * Don't override self.ordering with the output of get_ordering()
    * The filtering logic is significantly changed to deal with a tuple of fields.

    """

    def paginate_queryset(self, queryset, request, view=None):
        # This code only works with a QuerySetSequence.
        if not isinstance(queryset, QuerySetSequence):
            raise ValueError(
                "%s can only be used with an instance of QuerySetSequence." %
                self.__class__.__name__)

        self.page_size = self.get_page_size(request)
        if not self.page_size:
            return None

        self.base_url = request.build_absolute_uri()
        self.ordering = self.get_ordering(request, queryset, view)

        self.cursor = self.decode_cursor(request)
        if self.cursor is None:
            (offset, reverse, current_position) = (0, False, None)
        else:
            (offset, reverse, current_position) = self.cursor

        # Cursor pagination always enforces an ordering.
        if reverse:
            queryset = queryset.order_by(*_reverse_ordering(self.ordering))
        else:
            queryset = queryset.order_by(*self.ordering)

        # If we have a cursor with a fixed position then filter by that.
        if current_position is not None:
            # Iterate over each positioning.
            for order, position in zip(self.ordering[:len(current_position)], current_position):
                # The inside of this loop is essentially the old logic.
                is_reversed = order.startswith('-')
                order_attr = order.lstrip('-')

                # The filtering code in paginate_queryset uses ge or le, so
                # offset these based on that.
                if order_attr == '#':
                    equal = 'e'
                else:
                    equal = ''

                # Test for: (cursor reversed) XOR (queryset reversed)
                if self.cursor.reverse != is_reversed:
                    kwargs = {order_attr + '__lt' + equal: position}
                else:
                    kwargs = {order_attr + '__gt' + equal: position}

                # If filtering on the number of the QuerySet, apply it to the
                # entire QuerySetSequence.
                if order_attr == '#':
                    queryset = queryset.filter(**kwargs)

                # If there *are* QuerySets, filter just the edge QuerySet. This
                # avoids trimming items in subsequent QuerySets that are still
                # valid.
                elif queryset._querysets:
                    queryset = queryset._clone()

                    # Make a copy of the current QuerySets.
                    querysets = queryset._querysets
                    # Handle whether to look at the front edge or back edge of
                    # the QuerySets based on the order of iteration.
                    if self.cursor.reverse != is_reversed:
                        queryset._querysets = (
                            querysets[:-1] +
                            [querysets[-1].filter(**kwargs)])
                    else:
                        queryset._querysets = (
                            [querysets[0].filter(**kwargs)] +
                            querysets[1:])

        # If we have an offset cursor then offset the entire page by that amount.
        # We also always fetch an extra item in order to determine if there is a
        # page following on from this one.
        results = list(queryset[offset:offset + self.page_size + 1])
        self.page = list(results[:self.page_size])

        # Determine the position of the final item following the page.
        if len(results) > len(self.page):
            has_following_position = True
            following_position = self._get_position_from_instance(results[-1], self.ordering)
        else:
            has_following_position = False
            following_position = None

        # If we have a reverse queryset, then the query ordering was in reverse
        # so we need to reverse the items again before returning them to the user.
        if reverse:
            self.page = list(reversed(self.page))

        if reverse:
            # Determine next and previous positions for reverse cursors.
            self.has_next = (current_position is not None) or (offset > 0)
            self.has_previous = has_following_position
            if self.has_next:
                self.next_position = current_position
            if self.has_previous:
                self.previous_position = following_position
        else:
            # Determine next and previous positions for forward cursors.
            self.has_next = has_following_position
            self.has_previous = (current_position is not None) or (offset > 0)
            if self.has_next:
                self.next_position = following_position
            if self.has_previous:
                self.previous_position = current_position

        # Display page controls in the browsable API if there is more
        # than one page.
        if (self.has_previous or self.has_next) and self.template is not None:
            self.display_page_controls = True

        return self.page

    def get_ordering(self, *args, **kwargs):
        """Take whatever the expected ordering is and then first order by QuerySet."""
        result = super(SequenceCursorPagination, self).get_ordering(*args, **kwargs)

        # Because paginate_queryset sets self.ordering after reading it...we
        # need to only modify it sometimes. (This allows re-use of the
        # paginator, which probably only happens in tests.)
        if result[0] != '#':
            result = ('#', ) + result

        return result

    def _get_position_from_instance(self, instance, ordering):
        """
        The position will be a tuple of values:

            The QuerySet number inside of the QuerySetSequence.
            Whatever the normal value taken from the ordering property gives.

        """
        # Get the QuerySet number of the current instance.
        qs_order = getattr(instance, '#')

        # Strip the '#' and call the standard _get_position_from_instance.
        result = super(SequenceCursorPagination, self)._get_position_from_instance(instance, ordering[1:])

        # Return a tuple of these two elements.
        return (qs_order, result)

    def decode_cursor(self, request):
        """
        Given a request with a cursor, return a `Cursor` instance.

        Differs from the standard CursorPagination to handle a tuple in the
        position field.
        """
        # Determine if we have a cursor, and if so then decode it.
        encoded = request.query_params.get(self.cursor_query_param)
        if encoded is None:
            return None

        try:
            querystring = b64decode(encoded.encode('ascii')).decode('ascii')
            tokens = urlparse.parse_qs(querystring, keep_blank_values=True)

            offset = tokens.get('o', ['0'])[0]
            offset = _positive_int(offset, cutoff=self.offset_cutoff)

            reverse = tokens.get('r', ['0'])[0]
            reverse = bool(int(reverse))

            # The difference. Don't get just the 0th entry: get all entries.
            position = tokens.get('p', None)
        except (TypeError, ValueError):
            raise NotFound(self.invalid_cursor_message)

        return Cursor(offset=offset, reverse=reverse, position=position)



"""Test case for autocomplete implementations."""

import os
import uuid

from django import VERSION
from django.contrib.contenttypes.models import ContentType
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.db import transaction
try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse
from django.utils import six

from splinter import Browser


GLOBAL_BROWSER = None


class AutocompleteTestCase(StaticLiveServerTestCase):
    """Provide a class-persistent selenium instance and assertions."""

    @classmethod
    def setUpClass(cls):
        """Instanciate a browser for the whole test session."""
        global GLOBAL_BROWSER

        if GLOBAL_BROWSER is None:
            GLOBAL_BROWSER = Browser(os.environ.get('BROWSER', 'firefox'))
        cls.browser = GLOBAL_BROWSER

        super(AutocompleteTestCase, cls).setUpClass()

    def get(self, url):
        """Open a URL."""
        self.browser.visit('%s%s' % (
            self.live_server_url,
            url
        ))

        if '/admin/login/' in self.browser.url:
            # Should be pre-filled by HTML template
            # self.browser.fill('username', 'test')
            # self.browser.fill('password', 'test')
            self.browser.find_by_value('Log in').first.click()

        self.wait_script()

    def click(self, selector):
        """Click an element by css selector."""
        self.browser.find_by_css(selector).first.click()

    def enter_text(self, selector, text):
        """Enter text in an element by css selector."""
        self.browser.find_by_css(selector).first.value = ''
        self.browser.find_by_css(selector).first.type(text)

    def assert_not_visible(self, selector):
        """Assert an element is not visible by css selector."""
        e = self.browser.find_by_css(selector)
        assert not e or e.first.visible is False

    def assert_visible(self, selector):
        """Assert an element is visible by css selector."""
        e = self.browser.find_by_css(selector).first
        assert e.visible is True


class AdminMixin(object):
    """Mixin for tests that should happen in ModelAdmin."""

    def get_modeladmin_url(self, action, **kwargs):
        """Return a modeladmin url for a model and action."""
        return reverse('admin:%s_%s_%s' % (
            self.model._meta.app_label,
            self.model._meta.model_name,
            action
        ), kwargs=kwargs)

    def fill_name(self):
        """Fill in the name input."""
        i = self.id()
        half = int(len(i))
        not_id = i[half:] + i[:half]
        self.browser.fill('name', not_id)


class OptionMixin(object):
    """Mixin to make a unique option per test."""

    @transaction.atomic
    def create_option(self):
        """Create a unique option from self.model into self.option."""
        unique_name = six.text_type(uuid.uuid1())

        if VERSION < (1, 10):
            # Support for the name to be changed through a popup in the admin.
            unique_name = unique_name.replace('-', '')

        option, created = self.model.objects.get_or_create(
            name=unique_name)
        return option


class ContentTypeOptionMixin(OptionMixin):
    """Same as option mixin, with content type."""

    def create_option(self):
        """Return option, content type."""
        option = super(ContentTypeOptionMixin, self).create_option()
        ctype = ContentType.objects.get_for_model(option)
        return option, ctype



# All this craziness is so that we can allow the classes in nested_admin.formsets
# to be importable directly from this module, e.g.:
#
#     from nested_admin import NestedInlineFormSet
#
# without running afoul of the strict import order required by Django 1.9+.
# This implementation is shamelessly stolen from werkzeug's ``__init__.py``.
#
# Also included is a monkey-patch for django.forms.formsets.all_valid().
import pkg_resources
import sys
from types import ModuleType

import django
import django.forms.formsets
import monkeybiz

try:
    __version__ = pkg_resources.get_distribution('django-nested-admin').version
except pkg_resources.DistributionNotFound:
    __version__ = None

# import mapping to objects in other modules
all_by_module = {
    'nested_admin.formsets': (
        'NestedInlineFormSet', 'NestedBaseGenericInlineFormSet'),
    'nested_admin.nested': (
        'NestedModelAdmin', 'NestedModelAdminMixin', 'NestedInlineAdminFormset',
        'NestedInlineModelAdmin', 'NestedStackedInline', 'NestedTabularInline',
        'NestedInlineModelAdminMixin', 'NestedGenericInlineModelAdmin',
        'NestedGenericStackedInline', 'NestedGenericTabularInline')
}

# modules that should be imported when accessed as attributes of nested_admin
attribute_modules = frozenset(['formsets', 'nested'])

object_origins = {}
for module, items in all_by_module.items():
    for item in items:
        object_origins[item] = module


class module(ModuleType):

    def __dir__(self):
        """Just show what we want to show."""
        result = list(new_module.__all__)
        result.extend(('__file__', '__path__', '__doc__', '__all__',
                       '__docformat__', '__name__', '__path__',
                       '__package__', '__version__'))
        return result

    def __getattr__(self, name):
        if name in object_origins:
            module = __import__(object_origins[name], None, None, [name])
            for extra_name in all_by_module[module.__name__]:
                setattr(self, extra_name, getattr(module, extra_name))
            return getattr(module, name)
        elif name in attribute_modules:
            __import__('nested_admin.' + name)
        return ModuleType.__getattribute__(self, name)


# keep a reference to this module so that it's not garbage collected
old_module = sys.modules[__name__]

# setup the new module and patch it into the dict of loaded modules
new_module = sys.modules[__name__] = module(__name__)
new_module.__dict__.update({
    '__file__':         __file__,
    '__package__':      'nested_admin',
    '__path__':         __path__,
    '__doc__':          __doc__,
    '__version__':      __version__,
    '__all__':          tuple(object_origins) + tuple(attribute_modules),
    '__docformat__':    'restructuredtext en',
})

all_valid_patch_modules = [django.forms.formsets]

# If django.contrib.admin.options has already been imported, we'll need to
# monkeypatch all_valid in that module as well
admin_module = sys.modules.get('django.contrib.admin.options')
if admin_module:
    all_valid_patch_modules.append(admin_module)


@monkeybiz.patch(all_valid_patch_modules)
def all_valid(original_all_valid, formsets):
    """
    Checks validation on formsets, then handles a case where an inline
    has new data but one of its parent forms is blank.

    This causes a bug when one of the parent forms has empty_permitted == True,
    which happens if it is an "extra" form in the formset and its index
    is >= the formset's min_num.
    """
    if not original_all_valid(formsets):
        return False

    for formset in formsets:
        if formset.has_changed() and getattr(formset, 'parent_form', None):
            parent_form = formset.parent_form

            while True:
                if parent_form.empty_permitted:
                    parent_form.empty_permitted = False
                    # Reset the validation errors
                    parent_form._errors = None
                    if not parent_form.instance.pk and not parent_form.has_changed():
                        # Force Django to try to save the instance,
                        # since we need it for the fk to work
                        changed_data = parent_form.fields.keys()
                        if django.VERSION > (1, 9):
                            parent_form.__dict__['changed_data'] = changed_data
                        else:
                            parent_form._changed_data = changed_data
                if not hasattr(parent_form, 'parent_formset'):
                    break
                parent_form.parent_formset._errors = None
                if not hasattr(parent_form.parent_formset, 'parent_form'):
                    break
                parent_form = parent_form.parent_formset.parent_form

    if not original_all_valid(formsets):
        return False

    return True
